from __future__ import annotations

import asyncio
from typing import Any, Literal

import aiohttp
import loguru
from aiohttp import ClientError


class AsyncHTTPClient:
    def __init__(
        self,
        session: aiohttp.ClientSession,
        logger: loguru.Logger,
        retries: int = 3,
        backoff_factor: float = 1.0,
    ) -> None:
        self._session = session
        self._logger = logger
        self.retries = retries
        self.backoff_factor = backoff_factor

    async def _retry_request(
        self,
        method: Literal["GET", "POST", "DELETE", "PUT", "PATCH"],
        url: str,
        **kwargs: Any,
    ) -> aiohttp.ClientResponse:
        ex = Exception()
        for attempt in range(1, self.retries + 1):
            try:
                response = await self._session.request(method, url, **kwargs)
                if response.status < 500:
                    return response
            except ClientError as e:
                ex = e
                if attempt < self.retries:
                    self._logger.debug(f"Try {attempt}, exception during HTTP request - {e}")
                    delay = self.backoff_factor * attempt
                    await asyncio.sleep(delay)
                else:
                    self._logger.error(f"Exception during HTTP request - {ex}")
                    raise e
        raise ex or Exception("Unknown error during HTTP request")

    async def get(self, url: str, **kwargs: Any) -> aiohttp.ClientResponse:
        response = await self._retry_request("GET", url, **kwargs)
        return response

    async def post(self, url: str, data: Any | None = None, **kwargs: Any) -> aiohttp.ClientResponse:
        response = await self._retry_request("POST", url, data=data, **kwargs)
        return response

    async def put(self, url: str, data: Any | None = None, **kwargs: Any) -> aiohttp.ClientResponse:
        response = await self._retry_request("PUT", url, data=data, **kwargs)
        return response

    async def delete(self, url: str, **kwargs: Any) -> aiohttp.ClientResponse:
        response = await self._retry_request("DELETE", url, **kwargs)
        return response



import asyncio
from collections.abc import AsyncIterator
from http import HTTPStatus
from io import BytesIO
from pathlib import Path
from typing import BinaryIO
from uuid import UUID

import aiohttp
import attr
from aiohttp_socks import ProxyConnector
from loguru import logger

from ptsandbox.models import (
    CheckHealthResponse,
    GetVersionResponse,
    SandboxAdvancedScanTaskRequest,
    SandboxBaseTaskResponse,
    SandboxCheckTaskRequest,
    SandboxCheckTaskResponse,
    SandboxException,
    SandboxFileNotFoundException,
    SandboxGetImagesResponse,
    SandboxKey,
    SandboxRescanTaskRequest,
    SandboxScanTaskRequest,
    SandboxScanURLTaskRequest,
    SandboxUploadScanFileResponse,
)
from ptsandbox.utils.async_http_client import AsyncHTTPClient


class SandboxApi:
    """
    Using raw queries to sandbox API
    """

    key: SandboxKey
    session: aiohttp.ClientSession
    default_timeout: aiohttp.ClientTimeout
    upload_semaphore: asyncio.Semaphore

    def __init__(
        self,
        key: SandboxKey,
        *,
        default_timeout: aiohttp.ClientTimeout,
        upload_semaphore_size: int | None = None,
        proxy: str | None = None,
    ) -> None:
        self.key = key
        self.default_timeout = default_timeout
        self.session = aiohttp.ClientSession(
            timeout=self.default_timeout,
            connector=(
                ProxyConnector.from_url(proxy, ssl=False)
                if proxy
                else aiohttp.TCPConnector(
                    ssl=False,
                    # i know this is strange, but aiodns => c-ares can't correctly resolve dns names
                    # https://github.com/c-ares/c-ares/issues/642
                    resolver=aiohttp.ThreadedResolver(),
                )
            ),
            headers={"X-Api-Key": key.key},
        )
        self.http_client = AsyncHTTPClient(self.session, logger=logger)

        self.upload_semaphore = asyncio.Semaphore(
            upload_semaphore_size if upload_semaphore_size else self.key.max_workers
        )

    async def _upload_bytes(self, file: BinaryIO) -> AsyncIterator[bytes]:
        while chunk := file.read(1024 * 1024):
            yield chunk

    async def upload_file(
        self,
        file: str | Path | bytes | BinaryIO,
        upload_timeout: float = 300,
    ) -> SandboxUploadScanFileResponse:
        """
        Uploads the file to the sandbox

        Args:
            file: a path in the form of a string, either a Path object or binary data
            upload_timeout: if a large enough file is being uploaded, increase timeout (in seconds).

        Returns:
            The link to the file in the temporary storage and the lifetime of this file are returned, or an exception is thrown.

        Raises:
            SandboxException: if incorrect arguments are passed (usually when ignoring type hints)
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        # update default timeout
        timeout = attr.evolve(self.default_timeout, total=upload_timeout)

        url = f"{self.key.url}/storage/uploadScanFile"

        async with self.upload_semaphore:
            match file:
                case str() | Path():
                    # we can't use aiofiles here, because aiohttp try use chunked encoding
                    # sandbox (or maybe aiohttp) can't correctly handle chunked encoding (i have no idea why, don't ask me)
                    # so we need this clunky code
                    with open(file, "rb") as fd:
                        response = await self.http_client.post(
                            url=url,
                            data=fd,
                            timeout=timeout,
                        )
                case bytes():
                    response = await self.http_client.post(
                        url=url,
                        data=file,
                        timeout=timeout,
                    )
                case BytesIO():
                    response = await self.http_client.post(
                        url=url,
                        data=self._upload_bytes(file),
                        timeout=timeout,
                    )
                case _:
                    raise SandboxException(f"Specified file type doesn't supported {type(file)}!")

        response.raise_for_status()

        return await SandboxUploadScanFileResponse.build(response)

    async def _download_artifact(self, file_uri: str, read_timeout: int) -> aiohttp.ClientResponse:
        timeout = attr.evolve(self.default_timeout, sock_read=read_timeout)

        response = await self.http_client.post(
            f"{self.key.url}/storage/downloadArtifact",
            json={"file_uri": file_uri},
            timeout=timeout,
        )

        return response

    async def download_artifact(self, file_uri: str, read_timeout: int = 120) -> bytes:
        """
        Download file from the sandbox by hash

        Args:
            file_uri: id of the file in the sandbox
            read_timeout: how long should I wait for the file to download?

        Returns:
            File data

        Raises:
            SandboxFileNotFoundException: if the requested file is not found on the server
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        response = await self._download_artifact(file_uri, read_timeout)
        if response.status == HTTPStatus.NOT_FOUND:
            raise SandboxFileNotFoundException(f"Requested file {file_uri} not found")

        response.raise_for_status()

        # idk how to fix mypy complains about next line
        return await response.read()  # type: ignore

    async def download_artifact_stream(self, file_uri: str, read_timeout: int = 120) -> AsyncIterator[bytes]:
        """
        Download file from the sandbox by hash

        Args:
            file_uri: id of the file in the sandbox
            read_timeout: how long should I wait for the file to download?

        Returns:
            streaming file data or throw exception SandboxFileNotFoundException

        Raises:
            SandboxFileNotFoundException: if the requested file is not found on the server
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        response = await self._download_artifact(file_uri, read_timeout)
        if response.status == HTTPStatus.NOT_FOUND:
            raise SandboxFileNotFoundException(f"Requested file {file_uri} not found")

        response.raise_for_status()

        async for chunk in response.content.iter_chunked(1024 * 1024):
            yield chunk

    async def create_scan(self, data: SandboxScanTaskRequest, read_timeout: int = 0) -> SandboxBaseTaskResponse:
        """
        Send the specified file to the sandbox for analysis

        Args:
            data: sandbox parameters in model
            read_timeout: response waiting time in seconds

        Returns:
            The response from the sandbox is either with partial information (when using async_result), or with full information.

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        # compute default timeout
        timeout = attr.evolve(
            self.default_timeout,
            sock_read=data.options.sandbox.analysis_duration * 4
            + (300 if data.options.sandbox.analysis_duration < 80 else 120)
            + read_timeout,
        )

        response = await self.http_client.post(
            f"{self.key.url}/analysis/createScanTask",
            json=data.dict(),
            timeout=timeout,
        )

        response.raise_for_status()

        return await SandboxBaseTaskResponse.build(response)

    async def create_advanced_scan(
        self,
        data: SandboxAdvancedScanTaskRequest,
        read_timeout: int = 0,
    ) -> SandboxBaseTaskResponse:
        """
        Send the specified file to the sandbox for analysis using advanced APi

        Args:
            data: sandbox parameters in model
            read_timeout: response waiting time in seconds

        Returns:
            The response from the sandbox is either with partial information (when using async_result), or with full information.

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        # compute default timeout
        timeout = attr.evolve(
            self.default_timeout,
            sock_read=data.sandbox.analysis_duration * 4
            + (300 if data.sandbox.analysis_duration < 80 else 120)
            + read_timeout,
        )

        response = await self.http_client.post(
            f"{self.key.debug_url}/analysis/createBAScanTask",
            json=data.dict(),
            timeout=timeout,
        )

        response.raise_for_status()

        return await SandboxBaseTaskResponse.build(response)

    async def creat_url_scan(self, data: SandboxScanURLTaskRequest, read_timeout: int = 0) -> SandboxBaseTaskResponse:
        """
        Send the url to the sandbox

        Args:
            data: sandbox parameters in model
            read_timeout: response waiting time in seconds

        Returns:
            The response from the sandbox is either with partial information (when using async_result), or with full information.

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        timeout = attr.evolve(
            self.default_timeout,
            sock_read=data.options.sandbox.analysis_duration * 4
            + (300 if data.options.sandbox.analysis_duration < 80 else 120)
            + read_timeout,
        )

        response = await self.http_client.post(
            f"{self.key.url}/analysis/createScanURLTask",
            json=data.dict(),
            timeout=timeout,
        )

        response.raise_for_status()

        return await SandboxBaseTaskResponse.build(response)

    async def check_task(self, data: SandboxCheckTaskRequest) -> SandboxCheckTaskResponse:
        """
        Checking the result of a scan running with the async_result flag

        Args:
            task_id: task id :)
            allow_preflight:
                If this flag is set, an intermediate result with the `is_preflight` attribute
                will be returned for scanning with multiple stages (for example, static + BA).

        Returns:
            Information about the analysis status

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        response = await self.http_client.post(
            f"{self.key.url}/analysis/checkTask",
            json=data.dict(),
        )

        response.raise_for_status()

        return await SandboxCheckTaskResponse.build(response)

    async def get_report(self, scan_id: UUID) -> SandboxBaseTaskResponse:
        """
        Getting the full task scan report

        Args:
            task_id: task id :)

        Returns:
            The response from the sandbox is either with partial information (when using async_result), or with full information.

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        response = await self.http_client.post(
            f"{self.key.url}/analysis/report",
            json={"scan_id": str(scan_id)},
        )

        response.raise_for_status()

        return await SandboxBaseTaskResponse.build(response)

    async def create_rescan(self, data: SandboxRescanTaskRequest, read_timeout: int = 300) -> SandboxBaseTaskResponse:
        """
        Run a retro scan to check for detects without running a behavioral analysis.

        Args:
            data: sandbox parameters in model
            read_timeout: response waiting time in seconds

        Returns:
            The response from the sandbox is either with partial information (when using async_result), or with full information.

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        # compute default timeout
        timeout = attr.evolve(
            self.default_timeout,
            sock_read=(
                round(data.options.sandbox.analysis_duration * 1.5)
                if data.options.sandbox.analysis_duration > 70
                else 70
            )
            + read_timeout,
        )

        response = await self.http_client.post(
            f"{self.key.url}/analysis/createRetroTask",
            headers={"Content-Type": "application/json"},
            data=data.json(),
            timeout=timeout,
        )

        response.raise_for_status()

        return await SandboxBaseTaskResponse.build(response)

    async def get_email_headers(self, data: BinaryIO) -> AsyncIterator[bytes]:
        """
        Upload an email to receive headers

        Args:
            data: file data

        Returns:
            The header file

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        response = await self.http_client.post(
            f"{self.key.debug_url}/analysis/getHeaders",
            data=self._upload_bytes(data),
        )

        response.raise_for_status()

        async for chunk in response.content.iter_chunked(1024 * 1024):
            yield chunk

    async def get_images(self) -> SandboxGetImagesResponse:
        """
        Get a list of available images in the sandbox

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        response = await self.http_client.post(
            f"{self.key.url}/engines/sandbox/getImages",
            headers={"Content-Type": "application/json"},
        )

        response.raise_for_status()

        return await SandboxGetImagesResponse.build(response)

    async def check_health(self) -> CheckHealthResponse:
        """
        Checking the API status

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        response = await self.http_client.get(url=f"{self.key.url}/maintenance/checkHealth")

        response.raise_for_status()

        return await CheckHealthResponse.build(response)

    async def get_version(self) -> GetVersionResponse:
        """
        Get information about product

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        response = await self.http_client.get(url=f"{self.key.url}/maintenance/getVersion")

        response.raise_for_status()

        return await GetVersionResponse.build(response)

    def __del__(self) -> None:
        if not self.session.closed:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
            if loop.is_running():
                loop.create_task(self.session.close())
            else:
                loop.run_until_complete(self.session.close())



import asyncio
import datetime
import functools
import inspect
import random
from collections.abc import AsyncIterator, Awaitable, Callable
from http import HTTPStatus
from typing import Any, Literal, ParamSpec, TypeVar, overload
from urllib.parse import urlparse
from uuid import UUID

import aiohttp
import aiohttp.client_exceptions
from aiohttp_socks import ProxyConnector
from loguru import logger

from ptsandbox.models import (
    SandboxArtifactsFilterValuesResponse,
    SandboxAVEnginesResponse,
    SandboxBaqueueTasksResponse,
    SandboxClusterStatusResponse,
    SandboxCreateTokenResponse,
    SandboxException,
    SandboxKey,
    SandboxScansResponse,
    SandboxSystemSettingsResponse,
    SandboxSystemStatusResponse,
    SandboxSystemVersionResponse,
    SandboxTasksFilterValuesResponse,
    SandboxTasksResponse,
    SandboxTasksSummaryResponse,
    SandboxTokensResponse,
    SandboxTreeResponse,
    SandboxUpdateSystemSettingsRequest,
    TokenPermissions,
)
from ptsandbox.models.ui.components import SandboxComponentsResponse
from ptsandbox.models.ui.license import (
    SandboxLicenseResponse,
    SandboxLicenseUpdateResponse,
)
from ptsandbox.utils.async_http_client import AsyncHTTPClient

P = ParamSpec("P")
R = TypeVar("R")


class SandboxUI:
    """
    Using raw queries to sandbox UI API
    """

    const_headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/63.0.3239.132 Safari/537.36",  # noqa
        "Content-Type": "application/json",
    }

    key: SandboxKey
    session: aiohttp.ClientSession
    default_timeout: aiohttp.ClientTimeout

    update_token_lock: asyncio.Lock
    last_updated_token: datetime.datetime | None

    def __init__(
        self,
        key: SandboxKey,
        *,
        default_timeout: aiohttp.ClientTimeout,
        proxy: str | None = None,
        token_lifetime: datetime.timedelta = datetime.timedelta(minutes=8),
    ) -> None:
        self.key = key
        self.default_timeout = default_timeout
        self.token_lifetime = token_lifetime

        self.session = aiohttp.ClientSession(
            timeout=self.default_timeout,
            headers=self.const_headers,
            cookie_jar=aiohttp.CookieJar(unsafe=True),
            connector=(
                ProxyConnector.from_url(proxy, ssl=False)
                if proxy
                else aiohttp.TCPConnector(
                    ssl=False,
                    # i know this is strange, but aiodns => c-ares can't correctly resolve dns names
                    # https://github.com/c-ares/c-ares/issues/642
                    resolver=aiohttp.ThreadedResolver(),
                )
            ),
        )

        self.is_authorized = False
        self.last_updated_token = None
        self.fingerprint = "".join(random.choice("0123456789abcdef") for _ in range(32))
        self.update_token_lock = asyncio.Lock()

        self.http_client = AsyncHTTPClient(self.session, logger=logger)

    @staticmethod
    @overload
    def _token_required(func: Callable[P, Awaitable[R]]) -> Callable[P, Awaitable[R]]: ...

    @staticmethod
    @overload
    def _token_required(func: Callable[P, AsyncIterator[bytes]]) -> Callable[P, AsyncIterator[bytes]]: ...

    @staticmethod
    def _token_required(func: Callable[..., Any]) -> Callable[..., Any]:
        """
        An auxiliary method for verifying authorization
        """

        if inspect.isasyncgenfunction(func):

            @functools.wraps(func)
            # idk how to fix mypy complains about next line
            def wrapper_iter(self: "SandboxUI", *args: P.args, **kwargs: P.kwargs) -> Any:  # type: ignore
                async def inner() -> Any:
                    await self._ensure_token()
                    async for chunk in func(self, *args, **kwargs):
                        yield chunk

                return inner()

            return wrapper_iter
        else:

            @functools.wraps(func)
            # idk how to fix mypy complains about next line
            async def wrapper(self: "SandboxUI", *args: P.args, **kwargs: P.kwargs) -> Any:  # type: ignore
                await self._ensure_token()
                return await func(self, *args, **kwargs)

            return wrapper

    async def _ensure_token(self) -> None:
        if not self.is_authorized:
            raise SandboxException("Can't use the UI API without logging in first")

        async with self.update_token_lock:
            if not self.last_updated_token or (datetime.datetime.now() > self.last_updated_token + self.token_lifetime):
                await self._update_token()

    async def _update_token(self) -> None:
        response = await self.http_client.post(
            f"{self.key.ui_url}/auth/token",
            json={"fingerprint": self.fingerprint},
        )

        token = await response.json()

        try:
            self.session.headers["Authorization"] = "Bearer " + token["data"]["accessToken"]
        except ValueError as e:
            raise SandboxException("Can't get accessToken from response") from e

        self.last_updated_token = datetime.datetime.now()

    async def authorize(self) -> None:
        """
        Authorization in the UI using the passed parameters in the key
        """

        parameters = {"fingerprint": self.fingerprint}

        response = await self.http_client.get(f"{self.key.ui_url}/auth/authorize", params=parameters)
        try:
            location: str = (await response.json())["data"]["location"]
        except KeyError as e:
            raise SandboxException("Can't get location from authorization url") from e

        url = urlparse(location)

        assert self.key.ui is not None

        data: dict[str, str | bool | SandboxKey.UI.AuthType] = {
            "username": self.key.ui.login,
            "password": self.key.ui.password,
            "authType": self.key.ui.auth_type,
            "rememberLogin": True,
        }

        response = await self.http_client.post(f"{url.scheme}://{url.netloc}/ui/login", json=data)
        if response.status != HTTPStatus.OK:
            await self.close()
            response.raise_for_status()

        # get refresh token
        await self.http_client.get(location)

        self.is_authorized = True

        # get access token
        await self._update_token()

    async def close(self) -> None:
        """
        Close UI session
        """

        if not self.session.closed:
            await self.session.close()

    @_token_required
    async def get_system_status(self) -> SandboxSystemStatusResponse:
        """
        Get information about the system status

        For full information, look at the documentation of the `SandboxSystemStatusResponse` model.

        Returns:
            A model with information about the system
        """

        response = await self.http_client.get(f"{self.key.ui_url}/v2/system/status")

        response.raise_for_status()

        return SandboxSystemStatusResponse.model_validate(await response.json())

    @_token_required
    async def get_system_settings(self) -> SandboxSystemSettingsResponse:
        """
        Get information about the system settings

        Returns:
            A model with system settings
        """

        response = await self.http_client.get(f"{self.key.ui_url}/system/settings")

        response.raise_for_status()

        return SandboxSystemSettingsResponse.model_validate(await response.json())

    @_token_required
    async def update_system_settings(self, settings: SandboxUpdateSystemSettingsRequest) -> None:
        """
        Update system settings
        """

        response = await self.http_client.put(f"{self.key.ui_url}/system/settings", json=settings.dict())

        response.raise_for_status()

    @_token_required
    async def get_system_version(self) -> SandboxSystemVersionResponse:
        """
        Get the version of the installed product
        """

        response = await self.http_client.get(f"{self.key.ui_url}/system/version")

        response.raise_for_status()

        return SandboxSystemVersionResponse.model_validate(await response.json())

    @_token_required
    async def get_system_logs(
        self,
        since: int | None = None,
        components: list[str] | None = None,
    ) -> AsyncIterator[bytes]:
        """
        Download the archive with system logs

        Args:
            since: the time period for uploading logs in seconds
            components: component names in the format `{namespace}/{component}`, `{namespace}/{component}`, ... If the field is empty, all components will be downloaded

        Returns:
            Archive with logs
        """

        if components is None:
            components = []

        data: dict[str, Any] = {"components": ",".join(components)}
        if since is not None:
            data.update({"since": since})

        response = await self.http_client.get(f"{self.key.ui_url}/system/logs")

        response.raise_for_status()

        async for chunk in response.content.iter_chunked(1024 * 1024):
            yield chunk

    @_token_required
    async def get_system_cluster_status(self) -> SandboxClusterStatusResponse:
        """
        Get information about the cluster status
        """

        response = await self.http_client.get(f"{self.key.ui_url}/v2/system/status/cluster")

        response.raise_for_status()

        return SandboxClusterStatusResponse.model_validate(await response.json())

    @_token_required
    async def get_system_components_status(self) -> SandboxComponentsResponse:
        """
        Get information about the components status
        """

        response = await self.http_client.get(f"{self.key.ui_url}/v2/system/status/components")

        response.raise_for_status()

        return SandboxComponentsResponse.model_validate(await response.json())

    @_token_required
    async def get_license(self) -> SandboxLicenseResponse:
        """
        Get the license status and details
        """

        response = await self.http_client.get(f"{self.key.ui_url}/license")

        response.raise_for_status()

        return SandboxLicenseResponse.model_validate(await response.json())

    @_token_required
    async def update_license(self) -> SandboxLicenseUpdateResponse:
        """
        Updating the current license
        """

        response = await self.http_client.put(f"{self.key.ui_url}/license")

        response.raise_for_status()

        return SandboxLicenseUpdateResponse.model_validate(await response.json())

    @_token_required
    async def get_tasks(
        self,
        query: str = "",
        limit: int = 20,
        offset: int = 0,
        utc_offset_seconds: int = 0,
        next_cursor: str | None = None,
    ) -> SandboxTasksResponse:
        """
        Get tasks listing

        Args:
            query:
                filtering using the query language. For the syntax, see the user documentation.

                ```
                age < 30d AND (task.correlated.state != UNKNOWN ) ORDER BY start desc
                ```
            limit: limit on the number of records to be returned
            offset: the offset of the returned records. If the next Cursor is specified, the offset from the cursor is
            utc_offset_seconds: the offset of the user's time from UTC, which will be used for the time in QL queries


        Returns:
            Information about requested tasks
        """

        data: dict[str, Any] = {
            "query": query,
            "limit": limit,
            "offset": offset,
            "utcOffsetSeconds": utc_offset_seconds,
        }

        if next_cursor is not None:
            data.update({"nextCursor": next_cursor})

        response = await self.http_client.get(f"{self.key.ui_url}/v2/tasks", params=data)

        response.raise_for_status()

        return SandboxTasksResponse.model_validate(await response.json())

    @_token_required
    async def get_tasks_csv(
        self,
        query: str = "",
        columns: (
            list[
                Literal[
                    "action",
                    "behavioralAnalysis",
                    "fromTo",
                    "priority",
                    "processedTime",
                    "quarantine",
                    "source",
                    "status",
                    "taskName",
                    "time",
                    "verdict",
                    "verdictTime",
                ]
            ]
            | None
        ) = None,
        utc_offset_seconds: int = 0,
    ) -> AsyncIterator[bytes]:
        """
        Export a tasks listing to CSV

        Args:
            query: filtering using the query language. For the syntax, see the user documentation.
            columns: the list of csv columns to be exported.
            utc_offset_seconds: the offset of the user's time from UTC, which will be used for the time in QL queries

        Returns:
            AsyncIterator with chunks of CSV file
        """

        if columns is None:
            columns = []

        data: dict[str, Any] = {
            "format": "CSV",  # only csv supported by now
            "query": query,
            "columns": ",".join(columns),
            "utcOffsetSeconds": utc_offset_seconds,
        }

        response = await self.http_client.get(f"{self.key.ui_url}/v2/tasks/export", params=data)

        response.raise_for_status()

        async for chunk in response.content.iter_chunked(1024 * 1024):
            yield chunk

    @_token_required
    async def get_tasks_filter_values(
        self,
        from_: str = "",
        to: str = "",
        scan_id: UUID | None = None,
    ) -> SandboxTasksFilterValuesResponse:
        """
        Get possible values for filters based on sources and validation results

        Args:
            from_: for which period possible values are being searched: minimum time
            to: for which period possible values are being searched: maximum time
            scan_id: filter by task ID

        Returns:
            Possible filter values
        """

        data: dict[str, Any] = {}
        if scan_id is not None:
            data.update({"scanId": scan_id})

        if from_:
            data.update({"from": from_})

        if to:
            data.update({"to": to})

        response = await self.http_client.get(f"{self.key.ui_url}/v2/tasks/filter-values", params=data)

        response.raise_for_status()

        return SandboxTasksFilterValuesResponse.model_validate(await response.json())

    @_token_required
    async def get_task_summary(self, scan_id: UUID) -> SandboxTasksSummaryResponse:
        """
        Get information about a specific task

        Args:
            scan_id: task id

        Returns:
            Full information about a specific task
        """

        response = await self.http_client.get(f"{self.key.ui_url}/v2/tasks/{scan_id}/summary")

        response.raise_for_status()

        return SandboxTasksSummaryResponse.model_validate(await response.json())

    @_token_required
    async def get_artifacts_csv(
        self,
        query: str = "",
        columns: (
            list[
                Literal[
                    "behavioralAnalysis",
                    "bwListStatus",
                    "createProcess",
                    "detects.avast",
                    "detects.clamav",
                    "detects.drweb",
                    "detects.eset",
                    "detects.kaspersky",
                    "detects.nano",
                    "detects.ptesc",
                    "detects.vba",
                    "detects.yara",
                    "detects.yara.test",
                    "emlBcc",
                    "emlCC",
                    "emlFrom",
                    "emlTo",
                    "fileExtensionTypeGroup",
                    "fileLabels",
                    "fileMd5",
                    "fileName",
                    "fileSha1",
                    "fileSha256",
                    "fileSize",
                    "fileType",
                    "fromTo",
                    "imageDuration",
                    "imageName",
                    "mimeType",
                    "nodeType",
                    "priority",
                    "receivedFrom",
                    "ruleEngineDetects",
                    "ruleEngineVerdict",
                    "sandboxBehavioral",
                    "sandboxBootkitmon",
                    "sandboxDetects",
                    "sandboxVerdict",
                    "smtpFrom",
                    "smtpTo",
                    "source",
                    "ssdeep",
                    "status",
                    "subject",
                    "taskId",
                    "time",
                    "verdict",
                    "verdict.avast",
                    "verdict.clamav",
                    "verdict.drweb",
                    "verdict.eset",
                    "verdict.kaspersky",
                    "verdict.nano",
                    "verdict.ptesc",
                    "verdict.vba",
                    "verdict.yara",
                    "verdict.yara.test",
                    "verdictPriority",
                    "verdictReason",
                ]
            ]
            | None
        ) = None,
        utc_offset_seconds: int = 0,
    ) -> AsyncIterator[bytes]:
        """
        Export an artifacts listing to CSV

        Args:
            query: filtering using the query language. For the syntax, see the user documentation.
            columns: the list of csv columns to be exported.
            utc_offset_seconds: the offset of the user's time from UTC, which will be used for the time in QL queries

        Returns:
            AsyncIterator with chunks of CSV file
        """

        if columns is None:
            columns = []

        data: dict[str, Any] = {
            "format": "CSV",  # only csv supported by now
            "query": query,
            "columns": ",".join(columns),
            "utcOffsetSeconds": utc_offset_seconds,
        }

        response = await self.http_client.get(f"{self.key.ui_url}/v2/artifacts/export", params=data)

        response.raise_for_status()

        async for chunk in response.content.iter_chunked(1024 * 1024):
            yield chunk

    @_token_required
    async def get_artifacts_filter_values(
        self,
        from_: str = "",
        to: str = "",
        scan_id: UUID | None = None,
    ) -> SandboxArtifactsFilterValuesResponse:
        """
        Get possible values for filters based on sources and validation results

        Args:
            from_: for which period possible values are being searched: minimum time
            to: for which period possible values are being searched: maximum time
            scan_id: filter by task ID

        Returns:
            Possible filter values
        """

        data: dict[str, Any] = {}
        if scan_id is not None:
            data.update({"scanId": scan_id})

        if from_:
            data.update({"from": from_})

        if to:
            data.update({"to": to})

        response = await self.http_client.get(f"{self.key.ui_url}/v2/artifacts/filter-values", params=data)

        response.raise_for_status()

        return SandboxArtifactsFilterValuesResponse.model_validate(await response.json())

    @_token_required
    async def get_task_tree(
        self,
        scan_id: UUID,
        *,
        parent_path: list[int] | None = None,
        filtered_by_ids: list[int] | None = None,
        limit: int = 1000,
        offset: int = 0,
        max_tree_level: int = 3,
        sort_mode: Literal["DANGEROUS", "ALPHABETICAL"] = "DANGEROUS",
    ) -> SandboxTreeResponse:
        """
        Get a tree of artifacts for a specific task

        Args:
            scan_id: ...
            parent_path: the full path to the parent to start loading the tree from. For example: [0, 2, 10]
            filtered_by_ids: a list of IDs of specific nodes to be returned, for example: [0, 2, 10, 11]
            limit: limit on the number of records to be returned
            offset: the indentation from which the records are returned, used for pagination
            max_tree_level: the maximum depth (relative to the parent) to be returned
            sort_mode: the sorting method. First, the dangerous ones are 'DANGEROUS' or just alphabetically 'ALPHABETIC'

        Returns:
            The Artifact Tree
        """

        data: dict[str, Any] = {"limit": limit, "offset": offset, "maxTreeLevel": max_tree_level, "sortMode": sort_mode}
        if parent_path is not None:
            data.update({"parentPath": ",".join(map(str, parent_path))})
        if filtered_by_ids is not None:
            data.update({"filteredByIds": ",".join(map(str, filtered_by_ids))})

        response = await self.http_client.get(f"{self.key.ui_url}/v2/tasks/{scan_id}/tree", params=data)

        response.raise_for_status()

        return SandboxTreeResponse.model_validate(await response.json())

    @_token_required
    async def get_task_artifacts(
        self,
        scan_id: UUID,
        *,
        query: str = "",
        include_sandbox_logs: Literal["true", "false"] = "true",
        skip_data_files: Literal["true", "false"] = "false",
    ) -> AsyncIterator[bytes]:
        """
        Download all the artifacts of the task

        Args:
            scan_id: ...
            query: filtering using the query language. For the syntax, see the user documentation.
            include_sandbox_logs: whether to include BA logs as a result
            skip_data_files: whether to include data files in the result

        Returns:
            Sandbox returns an encrypted zip archive (password - infected), so we just export a set of bytes.
            If necessary, you can use pyzipper to unpack
        """

        data: dict[str, Any] = {
            "query": query,
            "includeSandboxLogs": include_sandbox_logs,
            "skip_data_files": skip_data_files,
        }

        response = await self.http_client.get(f"{self.key.ui_url}/v2/tasks/{scan_id}/tree/download", params=data)

        response.raise_for_status()

        async for chunk in response.content.iter_chunked(1024 * 1024):
            yield chunk

    @_token_required
    async def get_task_artifact_scans(self, scan_id: UUID, node_id: int) -> SandboxScansResponse:
        """
        Getting scan results for a specific artifact

        Args:
            scan_id: ...
            node_id: ...

        Returns:
            The model with the scan results
        """

        response = await self.http_client.get(f"{self.key.ui_url}/v2/tasks/{scan_id}/artifacts/{node_id}/scans")

        response.raise_for_status()

        return SandboxScansResponse.model_validate(await response.json())

    @_token_required
    async def get_baqueue_tasks(
        self,
        query: str = "age < 7d AND state IN (CREATED, STARTING, STARTED, DEDUPLICATION, READY, READY_WITH_ERROR) ORDER BY state DESC, priority.value DESC, ts.created",
        limit: int = 50,
        offset: int = 0,
        utc_offset_seconds: int = 0,
    ) -> SandboxBaqueueTasksResponse:
        """
        Listing of tasks in the Behavioral Analysis queue

        Args:
            query: QL search query (by default, all tasks that are currently running are requested)
            limit: limit on the number of records to be returned
            offset: offset of returned records
            utc_offset_seconds: the offset of the user's time from UTC, which will be used for the time in QL queries
        """

        data: dict[str, Any] = {
            "limit": limit,
            "offset": offset,
            "query": query,
            "utcOffsetSeconds": utc_offset_seconds,
        }

        response = await self.http_client.get(f"{self.key.ui_url}/baqueue/tasks", params=data)

        response.raise_for_status()

        return SandboxBaqueueTasksResponse.model_validate(await response.json())

    @_token_required
    async def get_av_engines(self) -> SandboxAVEnginesResponse:
        """
        Get information about antivirus scanners

        Returns:
            A model with information about all antiviruses
        """

        response = await self.http_client.get(f"{self.key.ui_url}/av-engines")

        response.raise_for_status()

        return SandboxAVEnginesResponse.model_validate(await response.json())

    async def get_api_tokens(self) -> SandboxTokensResponse:
        """
        Get listing of current Public API tokens

        Returns:
            A model with information about all tokens
        """

        response = await self.http_client.get(f"{self.key.ui_url}/public-api/tokens")

        response.raise_for_status()

        return SandboxTokensResponse.model_validate(await response.json())

    async def create_api_token(
        self,
        name: str,
        permissions: list[TokenPermissions],
        comment: str = "",
    ) -> SandboxCreateTokenResponse:
        """
        Create a new Public API token

        Args:
            name: token name
            permissions: permissions for the token
            comment: additional information about the token

        Returns:
            A model with information about the created token
        """

        response = await self.http_client.post(
            f"{self.key.ui_url}/public-api/tokens",
            json={
                "name": name,
                "permissions": permissions,
                "comment": comment,
            },
        )

        response.raise_for_status()

        return SandboxCreateTokenResponse.model_validate(await response.json())

    async def delete_api_token(self, token_id: int) -> None:
        """
        Delete the Public API token

        Args:
            token_id: id of the PublicAPI token in the database
        """

        response = await self.http_client.delete(f"{self.key.ui_url}/public-api/tokens/{token_id}")

        response.raise_for_status()

    def __del__(self) -> None:
        if not self.session.closed:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                loop = asyncio.new_event_loop()
            if loop.is_running():
                loop.create_task(self.close())
            else:
                loop.run_until_complete(self.close())



from pydantic import BaseModel, Field

from ptsandbox.models.ui.common import Scan


class SandboxScansResponse(BaseModel):
    """
    Scan results
    """

    scans: list[Scan] = []

    yara_test_info: Scan | None = Field(default=None, alias="yaraTestInfo")

    yara_main_info: Scan | None = Field(default=None, alias="yaraMainInfo")



from datetime import datetime
from uuid import UUID

from pydantic import BaseModel, Field

from ptsandbox.models.core import FileInfoTypes
from ptsandbox.models.ui.common import CorrelationInfo, EntryPoint, FilterValues


class Task(BaseModel):
    id: UUID
    """
    Task ID
    """

    name: str
    """
    Task Name
    """

    object_type: FileInfoTypes = Field(alias="objectType")
    """
    Object type
    """

    start: datetime
    """
    Task creation time (UNIX timestamp)
    """

    correlation: CorrelationInfo | None = None
    """
    The overall verdict of the product on the file. It's based on the sandbox,
    so it is on antiviruses and the result of static analysis.
    """

    sandbox_correlation: CorrelationInfo | None = Field(None, alias="sandboxCorrelation")
    """
    The verdict is exclusively sandbox
    """

    entry_point: EntryPoint = Field(alias="entryPoint")
    """
    Where did the task come from
    """

    start_time: float = Field(alias="startTime")
    """
    Task creation time (float UNIX timestamp)
    """

    processed_time: float = Field(alias="processedTime")
    """
    Task execution time (float UNIX timestamp)
    """

    verdict_time: float = Field(alias="verdictTime")
    """
    Task verdict time (float UNIX timestamp)
    """


class SandboxTasksResponse(BaseModel):
    """
    Tasks listing

    `<URL>/api/ui/v2/tasks`
    """

    tasks: list[Task]
    """
    Array of tasks
    """

    current_cursor: str = Field(alias="currentCursor")
    """
    The cursor for pagination, points to the data after the first record (if any)
    """

    next_cursor: str = Field(alias="nextCursor")
    """
    The cursor is for pagination, if the line is empty, then there is no more data.
    Indicates the data after the last record
    """


class SandboxTasksFilterValuesResponse(FilterValues):
    """
    Possible values for filters based on sources and validation results

    `<URL>/api/ui/v2/tasks/filter-values`
    """

    ...


class SandboxTasksSummaryResponse(BaseModel):
    """
    Information about a specific task

    `<URL>/api/ui/v2/tasks/{scanId}/summary`
    """

    task: Task



from pydantic import BaseModel

from ptsandbox.models.ui.common import Token


class SandboxTokensResponse(BaseModel):
    """
    Listing of current Public API tokens
    """

    total: int
    """
    The number of tokens in the system
    """

    entries: list[Token] = []
    """
    List of tokens
    """


class SandboxCreateTokenResponse(Token):
    token: str
    """
    The secret value of the token, which is shown only when creating a new PublicAPI token.
    """

    key: str
    """
    Hash of the secret value
    """



from pydantic import BaseModel, Field

from ptsandbox.models import ComponentStatus, ComponentType


class SandboxComponentsResponse(BaseModel):
    """
    Information about system components
    """

    class Component(BaseModel):
        class Pod(BaseModel):
            name: str

            component_name: str = Field(alias="componentName")
            """
            Which component does it belong to
            """

            ready: bool
            """
            Ready status
            """

            node: str
            """
            The name of the node it is running on
            """

            restarts: int

            error_reason: str = Field(alias="errorReason")
            """
            The type of error, if any
            """

            error_message: str = Field(alias="errorMessage")
            """
            Error message, if any
            """

            uptime: int
            """
            Time elapsed since the container was launched (in seconds)
            """

            containers_running: int = Field(alias="containersRunning")
            """
            The number of working containers for a given hearth
            """

            containers_total: int = Field(alias="containersTotal")
            """
            The total number of containers specified in the pod specification (excluding init containers)
            """

        name: str

        total_pods: int = Field(alias="totalPods")
        """
        How many pods are there in total
        """

        ready_pods: int = Field(alias="readyPods")
        """
        How many are running
        """

        nodes: list[str]
        """
        The list of nodes running the component's pods
        """

        status: ComponentStatus
        """
        Component status
        """

        type: ComponentType
        """
        Component type
        """

        pods: list[Pod] = []
        """
        Список подов
        """

    components: list[Component] = []



from pydantic import BaseModel

from ptsandbox.models.core.base import BaseResponse


class CheckHealthResponse(BaseResponse):
    """
    Healthcheck results
    """

    class Data(BaseModel):
        status: str
        """
        Health status
        """

    data: Data


class GetVersionResponse(BaseResponse):
    """
    Get information about product
    """

    class Data(BaseModel):
        version: str
        """
        Product version, for example '5.11.0.12345'
        """

        edition: str
        """
        Filled in for test builds or certification builds.
        """

    data: Data



import asyncio
import math
from collections.abc import AsyncIterator
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import BinaryIO
from uuid import UUID

import aiohttp
import aiohttp.client_exceptions
from aiohttp import ClientTimeout
from loguru import logger

from ptsandbox import config
from ptsandbox.models import (
    CheckHealthResponse,
    GetVersionResponse,
    SandboxAdvancedScanTaskRequest,
    SandboxBaseScanTaskRequest,
    SandboxBaseTaskResponse,
    SandboxCheckTaskRequest,
    SandboxCheckTaskResponse,
    SandboxException,
    SandboxImageInfo,
    SandboxKey,
    SandboxOptionsAdvanced,
    SandboxRescanTaskRequest,
    SandboxScanTaskRequest,
    SandboxScanURLTaskRequest,
    SandboxTooManyErrorsException,
    SandboxUploadException,
    SandboxWaitTimeoutException,
)
from ptsandbox.sandbox.sandbox_api import SandboxApi
from ptsandbox.sandbox.sandbox_ui import SandboxUI


class Sandbox:
    """
    The main class describing interaction with the sandbox via the API
    """

    api: SandboxApi
    ui: SandboxUI

    def __init__(
        self,
        key: SandboxKey,
        *,
        default_timeout: ClientTimeout = ClientTimeout(
            total=None,
            connect=None,
            sock_read=120.0,
            sock_connect=40.0,
        ),
        upload_semaphore_size: int | None = None,
        proxy: str | None = None,
    ) -> None:
        self.api = SandboxApi(
            key,
            default_timeout=default_timeout,
            upload_semaphore_size=upload_semaphore_size,
            proxy=proxy,
        )

        if key.ui is not None:
            self.ui = SandboxUI(key, default_timeout=default_timeout)

    async def create_rescan(
        self,
        trace: str | Path | bytes | BytesIO,
        network: str | Path | bytes | BytesIO,
        /,
        *,
        rules: str | Path | bytes | BytesIO | None = None,
        priority: int = 3,
        short_result: bool = False,
        async_result: bool = True,
        read_timeout: int = 300,
        options: SandboxBaseScanTaskRequest.Options = SandboxBaseScanTaskRequest.Options(),
    ) -> SandboxBaseTaskResponse:
        """
        Run a retro scan to check for detects without running a behavioral analysis.

        It is useful if there is a trace from a malware that can't connect to C2C.

        Or is it necessary to check the new correlation rules on the same trace.

        Args:
            trace: path to drakvuf-trace.log.zst or just bytes
            network: path to tcpdump.pcap or just bytes
            rules: if you have compiled the rules, then you can rescan with them, rather than using the sandbox embedded inside
            priority: the priority of the task, between 1 and 4. The higher it is, the faster it will get to work
            short_result:
                Return only the overall result of the check.

                The parameter value is ignored (true is used) if the value of the `async_result` parameter is also `true`.
            async_result:
                Return only the scan_id.

                Enabling this option may be usefull to send async requests for file checking.

                You can receive full report in a separate request.
            read_timeout: response waiting time in seconds
            options: additional sandbox options

        Returns:
            The response from the sandbox is either with partial information (when using async_result), or with full information.

        Raises:
            SandboxUploadException: if an error occurred when uploading files to the server
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        try:
            async with asyncio.TaskGroup() as tg:
                task_dummy = tg.create_task(self.api.upload_file(file=config.FAKE_PDF))
                task_trace = tg.create_task(self.api.upload_file(file=trace))
                task_network = tg.create_task(self.api.upload_file(file=network))

                if rules is not None:
                    task_rules = tg.create_task(self.api.upload_file(file=rules))
                else:
                    task_rules = None
        except ExceptionGroup as e:
            raise SandboxUploadException("Can't upload files to server") from e

        uploaded_dummy = task_dummy.result()
        uploaded_trace = task_trace.result()
        uploaded_network = task_network.result()

        if task_rules is not None:
            uploaded_rules = task_rules.result()
            options.sandbox.debug_options["rules_url"] = uploaded_rules.data.file_uri

        scan = SandboxRescanTaskRequest(
            file_uri=uploaded_dummy.data.file_uri,
            file_name=config.FAKE_NAME,
            raw_events_uri=uploaded_trace.data.file_uri,
            raw_network_uri=uploaded_network.data.file_uri,
            short_result=short_result,
            async_result=async_result,
            priority=priority,
            options=options,
        )

        return await self.api.create_rescan(scan, read_timeout)

    async def create_scan(
        self,
        file: str | Path | bytes | BinaryIO,
        /,
        *,
        file_name: str | None = None,
        rules: str | Path | bytes | BytesIO | None = None,
        priority: int = 3,
        short_result: bool = False,
        async_result: bool = True,
        read_timeout: int = 300,
        upload_timeout: float = 300,
        options: SandboxBaseScanTaskRequest.Options = SandboxBaseScanTaskRequest.Options(),
    ) -> SandboxBaseTaskResponse:
        """
        Send the specified file to the sandbox for analysis

        Args:
            file: the file to be sent for analysis
            file_name:
                The name of the file to be checked, which will be displayed in the sandbox web interface.

                If possible, the name of the uploaded file will be taken as the default value.

                If not specified, the hash value of the file is calculated using the SHA—256 algorithm.
            rules: if you have compiled the rules, then you can scan with them, rather than using the sandbox embedded inside
            priority: the priority of the task, between 1 and 4. The higher it is, the faster it will get to work
            short_result:
                Return only the overall result of the check.

                The parameter value is ignored (true is used) if the value of the `async_result` parameter is also `true`.
            async_result:
                Return only the scan_id.

                Enabling this option may be usefull to send async requests for file checking.

                You can receive full report in a separate request.
            read_timeout: response waiting time in seconds
            upload_timeout: if a large enough file is being uploaded, increase timeout (in seconds).
            options: additional sandbox options

        Returns:
            The response from the sandbox is either with partial information (when using async_result), or with full information.

        Raises:
            SandboxUploadException: if an error occurred when uploading files to the server
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        upload_name: str | None = file_name
        if not upload_name:
            match file:
                case str() | PathLike():
                    upload_name = str(file)
                case _:
                    upload_name = None

        try:
            async with asyncio.TaskGroup() as tg:
                task_file = tg.create_task(self.api.upload_file(file=file, upload_timeout=upload_timeout))
                if rules is not None:
                    task_rules = tg.create_task(self.api.upload_file(file=rules, upload_timeout=upload_timeout))
                else:
                    task_rules = None
        except ExceptionGroup as e:
            raise SandboxUploadException("Can't upload files to server") from e

        uploaded_file = task_file.result()

        if task_rules is not None:
            uploaded_rules = task_rules.result()
            options.sandbox.debug_options["rules_url"] = uploaded_rules.data.file_uri

        scan = SandboxScanTaskRequest(
            file_uri=uploaded_file.data.file_uri,
            file_name=upload_name,
            short_result=short_result,
            async_result=async_result,
            priority=priority,
            options=options,
        )

        return await self.api.create_scan(scan, read_timeout)

    async def create_advanced_scan(
        self,
        file: str | Path | bytes | BinaryIO,
        /,
        *,
        file_name: str | None = None,
        rules: str | Path | bytes | BytesIO | None = None,
        extra_files: list[Path] | None = None,
        short_result: bool = False,
        async_result: bool = True,
        read_timeout: int = 300,
        upload_timeout: float = 300,
        priority: int = 3,
        sandbox: SandboxOptionsAdvanced = SandboxOptionsAdvanced(),
    ) -> SandboxBaseTaskResponse:
        """
        Send the specified file to the sandbox for analysis using advanced API

        :warning: It may not be available in older versions of the sandbox.

        Args:
            file: the file to be sent for analysis
            file_name:
                The name of the file to be checked, which will be displayed in the sandbox web interface.

                If possible, the name of the uploaded file will be taken as the default value.

                If not specified, the hash value of the file is calculated using the SHA—256 algorithm.
            rules: if you have compiled the rules, then you can scan with them, rather than using the sandbox embedded inside
            priority: the priority of the task, between 1 and 4. The higher it is, the faster it will get to work
            short_result:
                Return only the overall result of the check.

                The parameter value is ignored (true is used) if the value of the `async_result` parameter is also `true`.
            async_result:
                Return only the scan_id.

                Enabling this option may be usefull to send async requests for file checking.

                You can receive full report in a separate request.
            read_timeout: response waiting time in seconds
            upload_timeout: if a large enough file is being uploaded, increase timeout (in seconds).
            sandbox: additional sandbox options

        Returns:
            The response from the sandbox is either with partial information (when using async_result), or with full information.

        Raises:
            SandboxUploadException: if an error occurred when uploading files to the server
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        upload_name: str | None = file_name
        if not upload_name:
            match file:
                case str() | PathLike():
                    upload_name = str(file)
                case _:
                    upload_name = None

        try:
            async with asyncio.TaskGroup() as tg:
                task_file = tg.create_task(self.api.upload_file(file=file, upload_timeout=upload_timeout))
                if rules is not None:
                    task_rules = tg.create_task(self.api.upload_file(file=rules, upload_timeout=upload_timeout))
                else:
                    task_rules = None

                if extra_files is not None:
                    tasks_extra_files = {str(file): tg.create_task(self.api.upload_file(file)) for file in extra_files}
                else:
                    tasks_extra_files = None
        except ExceptionGroup as e:
            raise SandboxUploadException("Can't upload files to server") from e

        uploaded_file = task_file.result()

        if tasks_extra_files is not None:
            for name, task in tasks_extra_files.items():
                uri = task.result().data.file_uri
                sandbox.extra_files.append(SandboxOptionsAdvanced.ExtraFile(name=name, uri=uri))

        if task_rules is not None:
            uploaded_rules = task_rules.result()
            sandbox.debug_options["rules_url"] = uploaded_rules.data.file_uri

        scan = SandboxAdvancedScanTaskRequest(
            file_uri=uploaded_file.data.file_uri,
            file_name=upload_name,
            short_result=short_result,
            async_result=async_result,
            priority=priority,
            sandbox=sandbox,
        )

        return await self.api.create_advanced_scan(scan, read_timeout)

    async def create_url_scan(
        self,
        url: str,
        /,
        *,
        rules: str | Path | bytes | BytesIO | None = None,
        priority: int = 3,
        short_result: bool = False,
        async_result: bool = True,
        read_timeout: int = 300,
        options: SandboxBaseScanTaskRequest.Options = SandboxBaseScanTaskRequest.Options(),
    ) -> SandboxBaseTaskResponse:
        """
        Send the url to the sandbox

        Args:
            url: the url to be sent for analysis
            rules: if you have compiled the rules, then you can scan with them, rather than using the sandbox embedded inside
            priority: the priority of the task, between 1 and 4. The higher it is, the faster it will get to work
            short_result:
                Return only the overall result of the check.

                The parameter value is ignored (true is used) if the value of the `async_result` parameter is also `true`.
            async_result:
                Return only the scan_id.

                Enabling this option may be usefull to send async requests for file checking.

                You can receive full report in a separate request.
            read_timeout: response waiting time in seconds
            options: additional sandbox options

        Returns:
            The response from the sandbox is either with partial information (when using async_result), or with full information.

        Raises:
            SandboxUploadException: if an error occurred when uploading files to the server
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        if rules is not None:
            try:
                uploaded_rules = await self.api.upload_file(file=rules)
            except aiohttp.client_exceptions.ClientResponseError as e:
                raise SandboxUploadException("Can't upload rules to server") from e

            options.sandbox.debug_options["rules_url"] = uploaded_rules.data.file_uri

        scan = SandboxScanURLTaskRequest(
            url=url,
            file_name=None,  # will be excluded
            short_result=short_result,
            async_result=async_result,
            priority=priority,
            options=options,
        )

        return await self.api.creat_url_scan(scan, read_timeout)

    async def wait_for_report(
        self,
        base_report: SandboxBaseTaskResponse,
        wait_time: float = 120,
        *,
        error_limit: int = 3,
    ) -> SandboxBaseTaskResponse:
        """
        Waiting for a full response from the sandbox if the request was with the `async_result=True` flag

        Args:
            base_time:
                how many seconds should I wait?

                Example of a formula for calculating a parameter:

                ```python
                wait_time = options.sandbox.analysis_duration * 4 + (
                    300 if sandbox_options.sandbox.analysis_duration < 80 else 120
                )
                ```

        Returns:
            The response from the sandbox with full information.

        Raises:
            SandboxException: there is nothing to wait, because there is not even a short report
            SandboxTooManyErrorsException: if there are too many errors while waiting for the report
            SandboxWaitTimeoutException: if the time is exceeded, the specified waiting time is
        """

        short_report = base_report.get_short_report()
        if not short_report:
            raise SandboxException("There is nothing to wait, because there is not even a short report")

        # if already have long report, just return it
        if base_report.get_long_report() is not None:
            return base_report

        elapsed_time: float = 0

        # calculate the sleep time, because for tasks with a long waiting time,
        # it makes no sense to poll the sandbox frequently
        sleep_time = math.ceil(wait_time / 64)

        error_counter = 0
        while elapsed_time <= wait_time:
            try:
                check = await self.get_report(short_report.scan_id)
            except Exception as ex:
                error_counter += 1

                logger.warning(f"Maybe dead sandbox {ex!r}, {self=}, {short_report.scan_id=}")

                if error_counter >= error_limit:
                    raise SandboxTooManyErrorsException("Too many errors while waiting report") from ex

                continue

            # full report is available?
            if check.get_long_report():
                return check

            await asyncio.sleep(sleep_time)

            elapsed_time += sleep_time

        raise SandboxWaitTimeoutException("Waiting time exceeded")

    async def check_task(self, task_id: str | UUID, allow_preflight: bool = True) -> SandboxCheckTaskResponse:
        """
        Checking the result of a scan running with the async_result flag

        Args:
            task_id: task id :)
            allow_preflight:
                If this flag is set, an intermediate result with the `is_preflight` attribute
                will be returned for scanning with multiple stages (for example, static + BA).

        Returns:
            Information about the analysis status

        Raises:
            SandboxException: if the passed task_id is not in UUID format
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        try:
            scan_id = UUID(task_id) if isinstance(task_id, str) else task_id
        except ValueError as e:
            raise SandboxException(f"Incorrect value={task_id} for task_id, expected UUID") from e

        return await self.api.check_task(
            SandboxCheckTaskRequest(
                scan_id=scan_id,
                allow_preflight=allow_preflight,
            )
        )

    async def get_report(self, task_id: str | UUID) -> SandboxBaseTaskResponse:
        """
        Getting the full task scan report

        The check was completed successfully. The results are in the message body. If the scan result is not ready yet, the result and artifacts keys are missing.

        :warning: The results will be returned only for the key that the analysis was started with. Sandbox restrictions for now.

        Args:
            task_id: task id :)

        Returns:
            The response from the sandbox is either with partial information (when using async_result), or with full information.

        Raises:
            SandboxException: if the passed task_id is not in UUID format
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        try:
            scan_id = UUID(task_id) if isinstance(task_id, str) else task_id
        except ValueError as e:
            raise SandboxException(f"Incorrect value={task_id} for task_id, expected UUID") from e

        return await self.api.get_report(scan_id=scan_id)

    def _get_hash_type(self, hash: str) -> str:
        """
        The simplest algorithm for determining the hash type
        """

        match len(hash):
            case 64:
                return "sha256"
            case _:
                raise SandboxException(f"Unknown hash type: {hash}")

    async def get_file(self, hash: str, read_timeout: int = 120) -> bytes:
        """
        Download file from the sandbox by hash

        Args:
            hash: sha256 hash of the file
            stream: download the entire file or give the result in chunks

        Returns:
            file data or throw exception SandboxFileNotFoundException

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        file_uri = f"{self._get_hash_type(hash)}:{hash}"
        return await self.api.download_artifact(file_uri=file_uri, read_timeout=read_timeout)

    async def get_file_stream(self, hash: str, read_timeout: int = 120) -> AsyncIterator[bytes]:
        """
        Download file from the sandbox by hash

        Args:
            hash: sha256 hash of the file
            stream: download the entire file or give the result in chunks

        Returns:
            file data or throw exception SandboxFileNotFoundException

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        file_uri = f"{self._get_hash_type(hash)}:{hash}"

        async for chunk in self.api.download_artifact_stream(file_uri=file_uri, read_timeout=read_timeout):
            yield chunk

    async def get_images(self) -> list[SandboxImageInfo]:
        """
        Get a list of available images in the sandbox

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        data = await self.api.get_images()
        return data.data

    async def get_email_headers(self, file: str | Path | bytes | BinaryIO) -> AsyncIterator[bytes]:
        """
        Upload an email to receive headers

        Args:
            file: path to .eml file or just binary data

        Returns:
            The header file

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        match file:
            case str() | Path():
                with open(file, "rb") as fd:
                    data = BytesIO(fd.read())
                iterator = self.api.get_email_headers(data)
            case bytes():
                iterator = self.api.get_email_headers(BytesIO(file))
            case BinaryIO():
                iterator = self.api.get_email_headers(file)
            case _:
                raise SandboxException(f"Unsupported type: {type(file)}")

        async for chunk in iterator:
            yield chunk

    async def check_health(self) -> CheckHealthResponse:
        """
        Checking the API status

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        return await self.api.check_health()

    async def get_version(self) -> GetVersionResponse:
        """
        Get information about product

        Raises:
            aiohttp.client_exceptions.ClientResponseError: if the response from the server is not ok
        """

        return await self.api.get_version()



from ptsandbox.models.ui.common import FilterValues


class SandboxArtifactsFilterValuesResponse(FilterValues):
    """
    Possible values for filters based on sources and validation results

    `<URL>/api/ui/v2/tasks/filter-values`
    """

    ...



from ptsandbox.models.ui.artifacts import SandboxArtifactsFilterValuesResponse
from ptsandbox.models.ui.av_engines import SandboxAVEnginesResponse
from ptsandbox.models.ui.baqueue import SandboxBaqueueTasksResponse
from ptsandbox.models.ui.cluster import SandboxClusterStatusResponse
from ptsandbox.models.ui.common import (
    CorrelationInfo,
    DetectionUI,
    EntryPoint,
    Error,
    FilterValues,
    HTTPDescription,
    MailResult,
    Scan,
    SMTPDefaultRecord,
    Token,
)
from ptsandbox.models.ui.components import SandboxComponentsResponse
from ptsandbox.models.ui.license import (
    SandboxLicenseResponse,
    SandboxLicenseUpdateResponse,
)
from ptsandbox.models.ui.scans import SandboxScansResponse
from ptsandbox.models.ui.system import (
    SandboxSystemSettingsResponse,
    SandboxSystemStatusResponse,
    SandboxSystemVersionResponse,
    SandboxUpdateSystemSettingsRequest,
)
from ptsandbox.models.ui.tasks import (
    SandboxTasksFilterValuesResponse,
    SandboxTasksResponse,
    SandboxTasksSummaryResponse,
    Task,
)
from ptsandbox.models.ui.tokens import SandboxCreateTokenResponse, SandboxTokensResponse
from ptsandbox.models.ui.tree import (
    SandboxInfo,
    SandboxTreeResponse,
    ScanArtifact,
    TreeNode,
)

__all__ = [
    "CorrelationInfo",
    "DetectionUI",
    "EntryPoint",
    "Error",
    "FilterValues",
    "HTTPDescription",
    "MailResult",
    "SMTPDefaultRecord",
    "SandboxAVEnginesResponse",
    "SandboxArtifactsFilterValuesResponse",
    "SandboxBaqueueTasksResponse",
    "SandboxClusterStatusResponse",
    "SandboxComponentsResponse",
    "SandboxCreateTokenResponse",
    "SandboxInfo",
    "SandboxLicenseResponse",
    "SandboxLicenseUpdateResponse",
    "SandboxScansResponse",
    "SandboxSystemSettingsResponse",
    "SandboxSystemStatusResponse",
    "SandboxSystemVersionResponse",
    "SandboxTasksFilterValuesResponse",
    "SandboxTasksResponse",
    "SandboxTasksSummaryResponse",
    "SandboxTokensResponse",
    "SandboxTreeResponse",
    "SandboxUpdateSystemSettingsRequest",
    "Scan",
    "ScanArtifact",
    "Task",
    "Token",
    "TreeNode",
]



from pydantic import BaseModel, Field


class SandboxAVEnginesResponse(BaseModel):
    class Data(BaseModel):
        class EnginesInfo(BaseModel):
            class Engine(BaseModel):
                class Error(BaseModel):
                    code: str
                    message: str

                distribution_type: str = Field(alias="distributionType")
                """
                Distribution type
                """

                engine_update_time: int = Field(alias="engineUpdateTime")
                """
                The time of the last update of the antivirus engine
                """

                distribution_pack: str = Field(alias="distributionPack")

                distribution_version: str = Field(alias="distributionVersion")
                """
                Distribution version
                """

                is_installed: bool = Field(alias="isInstalled")
                """
                Antivirus is installed
                """

                engine_version: str = Field(alias="engineVersion")
                """
                Engine version
                """

                enabled: bool
                """
                Antivirus is enabled
                """

                errors: list[Error] = []
                """
                Antivirus errors
                """

                is_initializing: bool = Field(alias="isInitializing")
                """
                Antivirus initialization status
                """

                is_ready: bool = Field(alias="isReady")
                """
                The antivirus is ready to work
                """

                database_time: int = Field(alias="databaseTime")
                """
                The time of the last database update
                """

                license_expiration: int = Field(alias="licenseExpiration")
                """
                License validity period
                """

                maintenance_status: str | None = Field(default=None, alias="maintenanceStatus")

            kaspersky: Engine | None = None
            bitdefender: Engine | None = None
            symantec: Engine | None = None
            eset: Engine | None = None
            drweb: Engine | None = None
            clamav: Engine | None = None
            avast: Engine | None = None
            avira: Engine | None = None

        engines_info: EnginesInfo = Field(alias="enginesInfo")

    data: Data


SandboxAVEnginesResponse.model_rebuild()



"""Модели, которые аналогичны и для API, и для UI"""

from datetime import datetime

from pydantic import AliasChoices, BaseModel, Field

from ptsandbox.models.core.base import BaseResponse
from ptsandbox.models.core.enum import (
    ArtifactType,
    EngineSubsystem,
    LogType,
    NetworkObjectType,
    SandboxImageType,
    ScanState,
    Verdict,
)


class SandboxResult(BaseModel):
    """
    File analysis result
    """

    scan_state: ScanState
    """
    Analysis status
    """

    duration: float
    """
    The duration of the analsysis in seconds.

    It is recorded only in the general results (in the JSON object `data → result`).
    """

    duration_full: float
    """
    The duration of the check, taking into account the record in the database or the time of the request.
    """

    verdict: Verdict | None = None
    """
    Analysis verdict
    """

    threat: str | None = None
    """
    Type of malware
    """

    errors: list[BaseResponse.Error] = []
    """
    Errors that occurred during analysis.
    """


class NetworkObject(BaseModel):
    type: NetworkObjectType
    """
    Type of network object
    """

    value: str
    """
    The value of the network object
    """


class SuspiciousBehaviors(BaseModel):
    """
    The correlation rule
    """

    name: str
    """
    Name of the rule
    """

    version: str | None = None
    """
    Version of the rule
    """

    mitre_threat_id: str = Field(
        validation_alias=AliasChoices(
            "mitre_threat_id",
            "mitreThreatId",
        )
    )
    """
    The MITRE Threat ID
    """

    weight: int
    """
    The weight of the rule (how much it affects the overall detection)
    """

    id: str | None = None
    description: str | None = None


class Detection(BaseModel):
    """
    Detected malware
    """

    detect: str
    """
    Malware
    """

    threat: str
    """
    Type of malware
    """


class Log(BaseModel):
    """
    A copy of network traffic, video recording, event logs, graph, debug files, mail headers.
    """

    type: LogType
    """
    Log type
    """

    file_uri: str
    """
    ID of the file used for downloading
    """

    file_name: str
    """
    File name
    """


class SandboxImageInfo(BaseModel):
    """
    Information about the VM image
    """

    class OS(BaseModel):
        """
        Information about the operating system of the virtual machine
        """

        name: str
        """
        Name of the operating system
        """

        version: str
        """
        Operating system version
        """

        architecture: str
        """
        Processor architecture supported by the operating system
        """

        service_pack: str | None = Field(
            default=None,
            validation_alias=AliasChoices("service_pack", "servicePack"),
        )
        """
        The name of the operating system update package
        """

        locale: str
        """
        Operating system locale
        """

    image_id: str | None = Field(default=None, validation_alias=AliasChoices("image_id", "name", "id"))
    """
    ID of the VM image

    The new UI began to return the name of the image. However, in the form of a name.
    """

    type: SandboxImageType | None = None
    """
    The type of image.
    """

    version: str | None = None
    """
    Version of the VM image
    """

    os: OS | None = None
    """
    Information about the operating system of the virtual machine image
    """


class Artifact(BaseModel):
    """
    A file, email, or other object
    """

    class FileInfo(BaseModel):
        """
        Information about the scanned file

        Filled in for binary objects
        """

        class FileInfoDetails(BaseModel):
            """
            The type of the nested object depends on the type value.
            """

            class ProcessDump(BaseModel):
                """
                It is filled in for process memory dumps.

                Type is equal to PROCESS_DUMP
                """

                process_name: str
                """
                The name of the process
                """

                process_id: int
                """
                The process ID (PID).
                """

                dump_trigger: str
                """
                The reason for extracting the dump
                """

                dump_create_time: float
                """
                Dump creation time
                """

            process_dump: ProcessDump

        file_uri: str
        """
        ID of the file used for downloading
        """

        file_path: str
        """
        The path to the file (excluding the root file of the structure), including its title.

        For example, for the file `readme.txt` at the root of the archive
        `archive.zip` will be specified as the value of this field.
        `readme.txt `, is an empty value for the archive itself.
        """

        mime_type: str
        """
        The MIME type of the artifact is determined during the verification process.
        """

        md5: str
        """
        MD5 hash of the file
        """

        sha1: str
        """
        SHA1 hash of the file
        """

        sha256: str
        """
        SHA256 hash of the file
        """

        ssdeep: str | None = None
        """
        SSDEEP hash of the file
        """

        size: int
        """
        File size in bytes
        """

        details: FileInfoDetails | None = None

    class EngineResult(BaseModel):
        class Details(BaseModel):
            class Sandbox(BaseModel):
                """
                Detailed information about behavioral analysis (if enabled)
                """

                class Stage(BaseModel):
                    """
                    The result of a single scan stage with bootkitmon
                    """

                    result: SandboxResult
                    """
                    The overall result of the check
                    """

                    detections: list[Detection] = []
                    """
                    A list of BA detections at this stage
                    """

                    logs: list[Log] = []
                    """
                    A copy of network traffic, video recording, event logs, graph, debug files, mail headers
                    """

                    artifacts: list["Artifact"] = []
                    """
                    Virtual machine artifacts are files created during behavioral analysis.
                    """

                    analysis_duration: float | None = None
                    """
                    The actual duration of the behavioral analysis in seconds
                    """

                    suspicious_behaviors: list[SuspiciousBehaviors] = []
                    """
                    List of triggered correlation rules
                    """

                image: SandboxImageInfo
                """
                Information about the VM image
                """

                logs: list[Log]
                """
                A copy of network traffic, video recording, event logs, graph, debug files, mail headers.
                """

                artifacts: list["Artifact"] | None = None
                """
                Virtual machine artifacts are files created during behavioral analysis.
                """

                stages: list[Stage] = []
                """
                The stages of bootkit analysis.
                """

                analysis_duration: float | None = None
                """
                The actual duration of the behavioral analysis in seconds
                """

                bootkitmon: bool | None = None
                """
                Was the bootkitmon analysis performed during BA
                """

                network_objects: list[NetworkObject] = []
                """
                Network objects (url, ip, domain)
                """

                suspicious_behaviors: list[SuspiciousBehaviors] = []
                """
                List of triggered correlation rules
                """

            sandbox: Sandbox | None = None
            """
            Detailed information about behavioral analysis (if enabled)
            """

        engine_subsystem: EngineSubsystem
        """
        The analysis method
        """

        engine_code_name: str
        """
        The name of the antivirus or component
        """

        engine_version: str | None = None
        """
        Antivirus or component version
        """

        database_version: str | None = None
        """
        Version of the antivirus database or knowledge base
        """

        database_time: datetime | None = None
        """
        Time to update the antivirus database or knowledge base
        """

        result: SandboxResult
        """
        The result of an antivirus or other component check
        """

        detections: list[Detection] = []
        """
        An array with a description of the detected malware
        """

        details: Details | None = None

    type: ArtifactType
    """
    The type of the analyzed object
    """

    result: SandboxResult | None = None
    """
    File analysis result
    """

    file_info: FileInfo | None = None
    """
    Information about the scanned file
    """

    engine_results: list[EngineResult] | None = None
    """
    The results of checking the file with specific antiviruses or other components
    """

    artifacts: list["Artifact"] | None = None
    """
    Files that are archived.

    If the file sent for analysis is not an archive or the allowed decompression depth is exceeded, the `artifacts` array is empty.
    """

    network_objects: list[NetworkObject]
    """
    Network objects (url, ip, domain)
    """

    def find_sandbox_result(self) -> EngineResult | None:
        if self.type == ArtifactType.ARCHIVE:
            if self.artifacts is not None:
                return self.artifacts[0].find_sandbox_result()
            return None
        if self.engine_results is not None:
            for result in self.engine_results:
                if result.engine_subsystem == EngineSubsystem.SANDBOX:
                    return result
        if self.artifacts is not None:
            for artifact in self.artifacts:
                if (sb_result := artifact.find_sandbox_result()) is not None:
                    return sb_result
        return None

    def find_static_result(self) -> EngineResult | None:
        if self.engine_results is not None:
            for result in self.engine_results:
                if result.engine_subsystem == EngineSubsystem.STATIC:
                    return result
        return None


# Solving the problem of nesting models into each other
Artifact.model_rebuild()
Artifact.EngineResult.Details.Sandbox.model_rebuild()



from pydantic import BaseModel

from ptsandbox.models.core.base import BaseResponse


class SandboxUploadScanFileResponse(BaseResponse):
    """
    Before running the file analysis using the API, your application must upload this file to the sandbox.

    `<URL>/storage/uploadScanFile`
    """

    class Data(BaseModel):
        file_uri: str
        """
        The ID of the uploaded file, used to create the analysis task.
        """

        ttl: int
        """
        The waiting time for the scan to start after uploading the file (in seconds).
        If the analysis has not been started during this time, the file will be deleted.
        """

    data: Data



from ptsandbox.models.api.analysis import (
    DebugOptions,
    SandboxAdvancedScanTaskRequest,
    SandboxBaseScanTaskRequest,
    SandboxBaseTaskResponse,
    SandboxCheckTaskRequest,
    SandboxCheckTaskResponse,
    SandboxOptions,
    SandboxOptionsAdvanced,
    SandboxRescanTaskRequest,
    SandboxScanTaskRequest,
    SandboxScanURLTaskRequest,
)
from ptsandbox.models.api.key import SandboxKey
from ptsandbox.models.api.maintenance import CheckHealthResponse, GetVersionResponse
from ptsandbox.models.api.sandbox import SandboxGetImagesResponse
from ptsandbox.models.api.storage import SandboxUploadScanFileResponse

__all__ = [
    "CheckHealthResponse",
    "DebugOptions",
    "GetVersionResponse",
    "SandboxAdvancedScanTaskRequest",
    "SandboxBaseScanTaskRequest",
    "SandboxBaseTaskResponse",
    "SandboxCheckTaskRequest",
    "SandboxCheckTaskResponse",
    "SandboxGetImagesResponse",
    "SandboxKey",
    "SandboxOptions",
    "SandboxOptionsAdvanced",
    "SandboxRescanTaskRequest",
    "SandboxScanTaskRequest",
    "SandboxScanURLTaskRequest",
    "SandboxUploadScanFileResponse",
]



"""
All enums are collected in one place, because there are many intersections in the models and it is easy to repeat.
"""

from enum import Enum
from typing import Any

from loguru import logger
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema


class SoftEnum(str, Enum):
    """
    A soft enum in order not to throw exceptions in production

    It is necessary because the library does not always keep up with api updates
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,
        handler: GetCoreSchemaHandler,  # pylint: disable=unused-argument
    ) -> CoreSchema:
        return core_schema.no_info_plain_validator_function(function=cls._validate)

    @classmethod
    def _validate(cls, value: str) -> "SoftEnum":
        if value not in cls.__members__.values():
            logger.warning(f'enum "{cls.__name__}" get unknown {value=}')

            # extended enum class with unknown value
            cls = Enum(
                cls.__name__,
                cls._member_map_ | {value.lower(): value},  # pylint: disable=self-cls-assignment,no-member
            )

        return cls(value)  # pyright: ignore[reportCallIssue, reportUnknownVariableType]


class HashType(SoftEnum):
    MD5 = "MD5"
    SHA1 = "SHA1"
    SHA256 = "SHA256"
    UNKNOWN = "UNKNOWN"


class EngineName(SoftEnum):
    CURL = "CURL"
    WEB_ENGINE = "WEB_ENGINE"


class ScanState(SoftEnum):
    """
    Status of completed analysis
    """

    UNKNOWN = "UNKNOWN"
    """
    What happened?
    """

    PARTIAL = "PARTIAL"
    """
    Partial check
    """

    FULL = "FULL"
    """
    Full check
    """

    UNSCANNED = "UNSCANNED"
    """
    The analysis was not carried out
    """


class Verdict(SoftEnum):
    CLEAN = "CLEAN"
    """
    No threats detected
    """

    UNWANTED = "UNWANTED"
    """
    Potentially dangerous
    """

    DANGEROUS = "DANGEROUS"
    """
    Malicious object
    """

    UNKNOWN = "UNKNOWN"
    """
    Threats are unknown (missing from the documentation)
    """


class ThreatPlatform(SoftEnum):
    ANDROID = "ANDROID"
    IOS = "IOS"
    LINUX = "LINUX"
    OSX = "OSX"
    WINDOWS = "WINDOWS"

    NO_PLATFORM = "NO_PLATFORM"


class ThreatClassification(SoftEnum):
    ADWARE = "ADWARE"
    BACKDOOR = "BACKDOOR"
    BOOTKIT = "BOOTKIT"
    CLIENT_IRC = "CLIENT_IRC"
    CLIENT_P2P = "CLIENT_P2P"
    CLIENT_SMTP = "CLIENT_SMTP"
    CONSTRUCTOR = "CONSTRUCTOR"
    DIALER = "DIALER"
    DOS = "DOS"
    DOWNLOADER = "DOWNLOADER"
    EMAIL_FLOODER = "EMAIL_FLOODER"
    EMAIL_WORM = "EMAIL_WORM"
    EXPLOIT = "EXPLOIT"
    FLOODER = "FLOODER"
    FRAUDTOOL = "FRAUDTOOL"
    HACKTOOL = "HACKTOOL"
    HOAX = "HOAX"
    IM_FLOODER = "IM_FLOODER"
    IM_WORM = "IM_WORM"
    IRC_WORM = "IRC_WORM"
    MONITOR = "MONITOR"
    NETTOOL = "NETTOOL"
    NET_WORM = "NET_WORM"
    P2P_WORM = "P2P_WORM"
    PHISHING = "PHISHING"
    PSWTOOL = "PSWTOOL"
    REMOTEADMIN = "REMOTEADMIN"
    RISKTOOL = "RISKTOOL"
    ROOTKIT = "ROOTKIT"
    SERVER_FTP = "SERVER_FTP"
    SERVER_PROXY = "SERVER_PROXY"
    SERVER_TELNET = "SERVER_TELNET"
    SERVER_WEB = "SERVER_WEB"
    SMS_FLOODER = "SMS_FLOODER"
    SPAM = "SPAM"
    SPOOFER = "SPOOFER"
    TROJAN = "TROJAN"
    TROJAN_ARCBOMB = "TROJAN_ARCBOMB"
    TROJAN_BANKER = "TROJAN_BANKER"
    TROJAN_CLICKER = "TROJAN_CLICKER"
    TROJAN_DDOS = "TROJAN_DDOS"
    TROJAN_DOWNLOADER = "TROJAN_DOWNLOADER"
    TROJAN_DROPPER = "TROJAN_DROPPER"
    TROJAN_FAKEAV = "TROJAN_FAKEAV"
    TROJAN_GAMETHIEF = "TROJAN_GAMETHIEF"
    TROJAN_IM = "TROJAN_IM"
    TROJAN_MAILFINDER = "TROJAN_MAILFINDER"
    TROJAN_NOTIFIER = "TROJAN_NOTIFIER"
    TROJAN_PROXY = "TROJAN_PROXY"
    TROJAN_PSW = "TROJAN_PSW"
    TROJAN_RANSOM = "TROJAN_RANSOM"
    TROJAN_SMS = "TROJAN_SMS"
    TROJAN_SPY = "TROJAN_SPY"
    UNKNOWN = "UNKNOWN"
    UNKNOWN_THREAT = "UNKNOWN_THREAT"
    VIRTOOL = "VIRTOOL"
    VIRUS = "VIRUS"
    WEBTOOLBAR = "WEBTOOLBAR"
    WORM = "WORM"


class VNCMode(SoftEnum):
    DISABLED = "DISABLED"
    """
    Manual analysis is disabled
    """

    FULL = "FULL"
    """
    Manual analysis is enabled
    """

    READ_ONLY = "READ_ONLY"
    """
    Manual analysis in viewing mode only
    """


class NetworkObjectType(SoftEnum):
    URL = "URL"
    IP = "IP"
    DOMAIN = "DOMAIN"


class LogType(SoftEnum):
    NETWORK = "NETWORK"
    """
    A copy of the network traffic in PCAP format
    """

    SCREENSHOT = "SCREENSHOT"
    """
    Snapshot or video recording from the virtual machine screen
    """

    EVENT_RAW = "EVENT_RAW"
    """
    Raw events from system
    """

    EVENT_CORRELATED = "EVENT_CORRELATED"
    """
    Correlated events
    """

    EVENT_NORMALIZED = "EVENT_NORMALIZED"
    """
    Normalized events
    """

    DEBUG = "DEBUG"
    """
    Debugging files
    """

    GRAPH = "GRAPH"
    """
    .graph file of the graph
    """


class ArtifactType(SoftEnum):
    """
    The type of the analyzed object
    """

    ARCHIVE = "ARCHIVE"
    COMPRESSED = "COMPRESSED"
    EMAIL = "EMAIL"
    FILE = "FILE"
    PROCESS_DUMP = "PROCESS_DUMP"
    URL = "URL"


class EngineSubsystem(SoftEnum):
    """
    The analysis method
    """

    AV = "AV"
    SANDBOX = "SANDBOX"
    STATIC = "STATIC"


class Action(SoftEnum):
    BLOCK = "BLOCK"
    NOTHING = "NOTHING"
    PASS = "PASS"
    UNKNOWN = "UNKNOWN"


class EmailType(SoftEnum):
    DISARMED = "DISARMED"
    NOTHING = "NOTHING"
    NOTIFICATION = "NOTIFICATION"
    SOURCE = "SOURCE"
    UNKNOWN = "UNKNOWN"


class DeliveryStatus(SoftEnum):
    FAIL = "FAIL"
    SKIP = "SKIP"
    SUCCESS = "SUCCESS"
    UNKNOWN = "UNKNOWN"


class EntryPointType(SoftEnum):
    CHECK_ME = "CHECK_ME"
    DPI = "DPI"
    FILE_INBOX = "FILE_INBOX"
    FILE_MONITOR = "FILE_MONITOR"
    ICAP = "ICAP"
    INTERACTIVE_ANALYSIS = "INTERACTIVE_ANALYSIS"
    MAIL_AGENT = "MAIL_AGENT"
    MAIL_BCC = "MAIL_BCC"
    MAIL_GATEWAY = "MAIL_GATEWAY"
    MAIL_GATEWAY_MTA = "MAIL_GATEWAY_MTA"
    PTNAD = "PTNAD"
    PT_CS = "PT_CS"
    PT_EDR = "PT_EDR"
    PUBLIC_API = "PUBLIC_API"
    SCAN_API = "SCAN_API"
    UNKNOWN = "UNKNOWN"
    WEB = "WEB"


class EntryPointStatus(SoftEnum):
    UNKNOWN = "UNKNOWN"
    SUCCESS = "SUCCESS"
    FAIL = "FAIL"


class EntryPointAction(SoftEnum):
    BLOCK = "BLOCK"
    NOTHING = "NOTHING"
    PASS = "PASS"
    UNKNOWN = "UNKNOWN"


class QuarantineState(SoftEnum):
    UNKNOWN = "UNKNOWN"
    QUARANTINED = "QUARANTINED"
    REMOVED = "REMOVED"


class QuarantineEventType(SoftEnum):
    QUARANTINE = "QUARANTINE"
    REMOVE = "REMOVE"
    SEND = "SEND"


class DPIState(SoftEnum):
    COMPLETED = "COMPLETED"
    ERROR = "ERROR"
    TRUNCATED = "TRUNCATED"
    UNKNOWN = "UNKNOWN"


class FileInfoTypes(SoftEnum):
    ARCHIVE = "ARCHIVE"
    COMPRESSED_FILE = "COMPRESSED_FILE"
    EMAIL = "EMAIL"
    EMAIL_BODY = "EMAIL_BODY"
    FILE = "FILE"
    FOLDER = "FOLDER"
    HTTP = "HTTP"
    SANDBOX_DROP = "SANDBOX_DROP"
    SANDBOX_MEMORY_DUMP = "SANDBOX_MEMORY_DUMP"
    SANDBOX_PROCESS_MEMORY_DUMP = "SANDBOX_PROCESS_MEMORY_DUMP"
    URL = "URL"


class FileInfoProperties(SoftEnum):
    ARCHIVE = "ARCHIVE"
    COMPRESSED = "COMPRESSED"
    CORRUPTED = "CORRUPTED"
    EMAIL = "EMAIL"
    ENCRYPTED = "ENCRYPTED"
    HAS_ACTION = "HAS_ACTION"
    HAS_ACTIVE_X = "HAS_ACTIVE_X"
    HAS_ADD_IN = "HAS_ADD_IN"
    HAS_DDE = "HAS_DDE"
    HAS_EMBEDDED = "HAS_EMBEDDED"
    HAS_JAVASCRIPT = "HAS_JAVASCRIPT"
    HAS_MACROS = "HAS_MACROS"
    HAS_OPEN_ACTION = "HAS_OPEN_ACTION"
    HAS_REMOTE_DATA = "HAS_REMOTE_DATA"
    HAS_REMOTE_TEMPLATE = "HAS_REMOTE_TEMPLATE"
    MULTI_VOLUME = "MULTI_VOLUME"
    OFFICE = "OFFICE"
    PROTECTED = "PROTECTED"
    PY_INSTALLER = "PY_INSTALLER"
    SFX = "SFX"
    SFX_7Z = "SFX_7z"
    SFX_7z = "SFX_7z"
    SFX_ACE = "SFX_ACE"
    SFX_RAR = "SFX_RAR"
    SFX_ZIP = "SFX_ZIP"
    UPX = "UPX"


class TreeEngineName(SoftEnum):
    BITDEFENDER = "bitdefender"
    CLAMAV = "clamav"
    DRWEB = "drweb"
    KASPERSKY = "kaspersky"
    NANO = "nano"
    PTESC = "ptesc"
    PTIOC = "ptioc"
    PT_CATEGORIZER = "ptcategorizer"
    PT_SANDBOX_OVERALL = "pt_sandbox_overall"
    RULE_ENGINE = "rule_engine"
    VBA = "vba"
    YARA_ENGINE = "yara_engine"


class TreeNodeType(SoftEnum):
    ARTIFACT = "ARTIFACT"
    SANDBOX = "SANDBOX"
    SANDBOX_STAGE = "SANDBOX_STAGE"


class ScanArtifactType(SoftEnum):
    CORRELATED = "SANDBOX_CORRELATED_EVENT"
    DEBUG = "SANDBOX_DEBUG_FILE"
    EMAIL = "EMAIL_HEADERS_PTESC"
    GRAPH = "SANDBOX_GRAPH"
    NORMALIZED = "SANDBOX_NORMALIZED_EVENT"
    PCAP = "SANDBOX_NETWORK_FILE"
    RAW_EVENT_FILE = "SANDBOX_RAW_EVENT_FILE"
    VIDEO = "SANDBOX_VIDEO"


class ContextType(SoftEnum):
    EMPTY = ""
    CRAWLER = "CRAWLER"
    PTESC = "PTESC"
    SANDBOX = "SANDBOX"


class BootkitmonStage(SoftEnum):
    UNKNOWN = "UNKNOWN"
    BEFORE_REBOOT = "BEFORE_REBOOT"
    AFTER_REBOOT = "AFTER_REBOOT"


class ErrorType(SoftEnum):
    BOOTKITMON_REBOOT_TIMEOUT = "BOOTKITMON_REBOOT_TIMEOUT"
    CANCELLED_BY_RULES = "CANCELLED_BY_RULES"
    CANCELLED_BY_USER = "CANCELLED_BY_USER"
    COLLISION_ERROR = "COLLISION_ERROR"
    CONNECTION_ERROR = "CONNECTION_ERROR"
    CONNECT_TIMEOUT = "CONNECT_TIMEOUT"
    CONTAINS_ENCRYPTED = "CONTAINS_ENCRYPTED"
    CORRUPTED = "CORRUPTED"
    ENCRYPTED = "ENCRYPTED"
    ENGINE_ERROR = "ENGINE_ERROR"
    FILE_NOT_FOUND = "FILE_NOT_FOUND"
    INIT_ERROR = "INIT_ERROR"
    LIMIT_EXCEEDED = "LIMIT_EXCEEDED"
    LISTS_NOT_READY_ERROR = "LISTS_NOT_READY_ERROR"
    MAX_DOWNLOAD_LIMIT_EXCEEDED = "MAX_DOWNLOAD_LIMIT_EXCEEDED"
    MAX_REDIRECT_EXCEEDED = "MAX_REDIRECT_EXCEEDED"
    MAX_SIZE_EXCEEDED = "MAX_SIZE_EXCEEDED"
    NODE_LIMIT_EXCEEDED = "NODE_LIMIT_EXCEEDED"
    NOT_ALLOWED_REDIRECT = "NOT_ALLOWED_REDIRECT"
    NOT_ENOUGH_IMAGE_COPIES = "NOT_ENOUGH_IMAGE_COPIES"
    NOT_FILE = "NOT_FILE"
    NOT_UNPACKABLE_FILE = "NOT_UNPACKABLE_FILE"
    NO_SUITABLE_UNPACKER = "NO_SUITABLE_UNPACKER"
    READ_TIMEOUT = "READ_TIMEOUT"
    RESPONSE_ERROR = "RESPONSE_ERROR"
    SCAN_MACHINE_ERROR = "SCAN_MACHINE_ERROR"
    TIMEOUT_ERROR = "TIMEOUT_ERROR"
    UNKNOWN = "UNKNOWN"


class HTTPDirection(SoftEnum):
    REQUEST = "REQUEST"
    RESPONSE = "RESPONSE"
    UNKNOWN = "UNKNOWN"


class BlacklistStatus(SoftEnum):
    IN_BLACK_LIST = "IN_BLACK_LIST"
    IN_WHITE_LIST = "IN_WHITE_LIST"
    NOT_IN_LISTS = "NOT_IN_LISTS"
    UNKNOWN = "UNKNOWN"


class SandboxImageType(SoftEnum):
    BASE = "base"
    CUSTOM = "custom"
    USER = "user"


class BaqueueState(SoftEnum):
    CREATED = "CREATED"
    DEDUPLICATION = "DEDUPLICATION"
    FINISHED_SUCCESSFULLY = "FINISHED_SUCCESSFULLY"
    FINISHED_WITH_ERROR = "FINISHED_WITH_ERROR"
    READY = "READY"
    READY_WITH_ERROR = "READY_WITH_ERROR"
    STARTED = "STARTED"
    STARTING = "STARTING"


class TokenPermissions(SoftEnum):
    SCAN_WITH_EXTENDED_SETTINGS = "SCAN_WITH_EXTENDED_SETTINGS"
    SCAN_WITH_PREDEFINED_SETTINGS = "SCAN_WITH_PREDEFINED_SETTINGS"


class SystemGroup(SoftEnum):
    AUTO_UPDATE = "AUTO_UPDATE"
    DATA_DB = "DATA_DB"
    ENGINE = "ENGINE"
    ENTRY_POINT = "ENTRY_POINT"
    EOS = "EOS"
    EVENTS_DB = "EVENTS_DB"
    FILES_STORAGE = "FILES_STORAGE"
    NODE = "NODE"
    SANDBOX = "SANDBOX"
    SERVICES = "SERVICES"


class SystemCode(SoftEnum):
    AV_ENGINE_IN_NOT_GENERAL_AVAILABILITY = "AV_ENGINE_IN_NOT_GENERAL_AVAILABILITY"
    COMPONENT_CAPACITY_EXCEEDED_MAX_SIZE = "COMPONENT_CAPACITY_EXCEEDED_MAX_SIZE"
    COMPONENT_CAPACITY_EXCEEDED_THRESHOLD = "COMPONENT_CAPACITY_EXCEEDED_THRESHOLD"
    COMPONENT_ERROR = "COMPONENT_ERROR"
    COMPONENT_PARTIALLY_AVAILABLE = "COMPONENT_PARTIALLY_AVAILABLE"
    END_OF_SUPPORT = "END_OF_SUPPORT"
    END_OF_SUPPORT_SOON = "END_OF_SUPPORT_SOON"
    IMAGE_INSTALL_ERROR = "IMAGE_INSTALL_ERROR"
    NEW_VERSION_AVAILABLE = "NEW_VERSION_AVAILABLE"
    NEW_VERSION_INSTALLATION_SCHEDULED = "NEW_VERSION_INSTALLATION_SCHEDULED"
    NODE_HAVE_ERROR = "NODE_HAVE_ERROR"
    NODE_IS_NOT_READY = "NODE_IS_NOT_READY"
    SANDBOX_RECONFIGURING = "SANDBOX_RECONFIGURING"
    STORAGE_CAPACITY_EXCEEDED_COMPONENTS_MAX_SIZE_BYTES = "STORAGE_CAPACITY_EXCEEDED_COMPONENTS_MAX_SIZE_BYTES"
    STORAGE_CAPACITY_EXCEEDED_MINIMUM = "STORAGE_CAPACITY_EXCEEDED_MINIMUM"


class UploadStrategy(SoftEnum):
    ALL_THREATS = "ALL_THREATS"
    DO_NOT_UPLOAD = "DO_NOT_UPLOAD"
    THREAT_SOURCE_ONLY = "THREAT_SOURCE_ONLY"


class EOSStatus(SoftEnum):
    ENDED = "ENDED"
    ENDS_SOON = "ENDS_SOON"
    OK = "OK"


class ComponentStatus(SoftEnum):
    NOT_READY = "NOT_READY"
    PARTIALLY_READY = "PARTIALLY_READY"
    READY = "READY"


class ComponentType(SoftEnum):
    BEHAVIORAL_ANALYSIS = "BEHAVIORAL_ANALYSIS"
    CORE = "CORE"
    DATABASE = "DATABASE"
    FILE_STORAGE = "FILE_STORAGE"
    INTEGRATIONS = "INTEGRATIONS"
    MANAGEMENT = "MANAGEMENT"
    MONITORING = "MONITORING"
    OTHER = "OTHER"
    SCAN_ENGINES = "SCAN_ENGINES"
    SOURCES = "SOURCES"
    USER = "USER"
    VERIFICATION_RESULTS = "VERIFICATION_RESULTS"


class LicenseStatus(SoftEnum):
    EXPIRED = "EXPIRED"
    NO_LICENSE = "NO_LICENSE"
    VALID = "VALID"


class LicenseUpdateError(SoftEnum):
    BAD_LICENSE = "BAD_LICENSE"
    FUS_UNAVAILABLE = "FUS_UNAVAILABLE"
    LICENSE_FROM_ANOTHER_PRODUCT = "LICENSE_FROM_ANOTHER_PRODUCT"
    LICENSE_NOT_FOUND = "LICENSE_NOT_FOUND"
    NO_ERROR = ""
    PRODUCT_NOT_SUPPORTED = "PRODUCT_NOT_SUPPORTED"
    TIMEOUT = "TIMEOUT"
    UNKNOWN_ERROR = "UNKNOWN_ERROR"


class LicenseEntryPoint(SoftEnum):
    CHECK_ME = "check-me"
    EXCHANGE_MTA = "exchange-mta"
    FILES_INBOX = "files-inbox"
    FILES_MONITOR = "files-monitor"
    ICAP = "icap"
    MAIL_BCC = "mail-bcc"
    MAIL_GATEWAY = "mail-gateway"
    MAIL_GATEWAY_MTA = "mail-gateway-mta"
    PT_CS = "pt-cs"
    PT_EDR = "pt-edr"
    PT_NAD = "pt-nad"
    PUBLIC_API = "public-api"
    SCAN_API = "scan-api"


class LicenseAvEngine(SoftEnum):
    AVAST = "avast"
    AVIRA = "avira"
    BITDEFENDER = "bitdefender"
    CLAMAV = "clamav"
    DRWEB = "drweb"
    ESET = "eset"
    KASPERSKY = "kaspersky"
    NANO = "nano"
    SYMANTEC = "symantec"
    VBA = "vba"


class LicensePerformanceType(SoftEnum):
    EMAIL_LIMIT = "EMAIL_LIMIT"
    NETWORK_STORAGE_LIMIT = "NETWORK_STORAGE_LIMIT"
    NETWORK_TRAFFIC_LIMIT = "NETWORK_TRAFFIC_LIMIT"



from typing import Literal

from pydantic import AwareDatetime, BaseModel, Field

from ptsandbox.models import (
    LicenseAvEngine,
    LicenseEntryPoint,
    LicensePerformanceType,
    LicenseStatus,
    LicenseUpdateError,
)


class SandboxLicenseResponse(BaseModel):
    """
    License status and details
    """

    class Data(BaseModel):
        class UpdateStatus(BaseModel):
            """
            Information about the last attempt to update the license
            """

            error: LicenseUpdateError
            """
            Error if the status is FAILED
            """

            last_check_status: Literal["SUCCESS", "FAILED"] = Field(alias="lastCheckStatus")

            last_check_time: AwareDatetime = Field(alias="lastCheckTime")

            last_success_check_time: AwareDatetime = Field(alias="lastSuccessCheckTime")

            license_update_time: AwareDatetime = Field(alias="licenseUpdateTime")
            """
            When the license itself was updated, not when it was checked
            """

            product: Literal["Sandbox", "MultiScanner"]
            """
            Product type
            """

        class License(BaseModel):
            class NodesLimit(BaseModel):
                multiscanner: int = Field(alias="multiScanner")
                sandbox_high_performance: int = Field(alias="sandboxHighPerformance")
                sandbox_low_performance: int = Field(alias="sandboxLowPerformance")

            class Performance(BaseModel):
                """
                Bandwidth by traffic type
                """

                type: LicensePerformanceType
                limit: int

            class Sandbox(BaseModel):
                enabled: bool
                """
                Is behavioral analysis allowed?
                """

                images: list[str]
                """
                Images that can be used in behavioral analysis
                """

            class Telemetry(BaseModel):
                enabled: bool
                """
                Is telemetry enabled
                """

            entry_points: list[LicenseEntryPoint] = Field(alias="entryPoints")
            """
            Allowed entrypoints
            """

            expiration_time: AwareDatetime = Field(alias="expirationTime")
            """
            License expiration date
            """

            external_av_engines: list[LicenseAvEngine] = Field(alias="externalAvEngines")
            """
            Allowed external engines
            """

            files_per_hour: int = Field(alias="filesPerHour")
            """
            Throughput capacity
            """

            grace_period: int = Field(alias="gracePeriod")
            """
            The number of grace period days in seconds
            """

            perpetual: bool
            """
            Is the license permanent
            """

            internal_av_engines: list[LicenseAvEngine] = Field(alias="internalAvEngines")
            """
            Allowed internal engines
            """

            is_entry_points_blocking_mode: bool = Field(alias="isEntryPointsBlockingMode")
            """
            Is blocking mode allowed?
            """

            license_version: int = Field(default=2, alias="licenseVersion")
            """
            License version
            """

            nodes_limit: NodesLimit = Field(alias="nodesLimit")
            """
            Maximum number of nodes by type
            """

            number: int
            """
            License number issued
            """

            performance: list[Performance] = []
            """
            Bandwidth by traffic type
            """

            sandbox: Sandbox

            telemetry: Telemetry

        state: LicenseStatus
        """
        License status - is there, is it expired
        """

        serial_number: str = Field(alias="serialNumber")
        """
        Serial number of the current license
        """

        update_status: UpdateStatus = Field(alias="updateStatus")

        license: License

    data: Data


class SandboxLicenseUpdateResponse(BaseModel):
    """
    License update attempt status
    """

    class Data(BaseModel):
        status: Literal["SUCCESS", "FAILED"]

        error: LicenseUpdateError

    data: Data



from ptsandbox.models.core.base import BaseResponse
from ptsandbox.models.core.common import SandboxImageInfo


class SandboxGetImagesResponse(BaseResponse):
    """
    Your application can get a list of virtual machine images installed in the PT Sandbox.

    `<URL>/engines/sandbox/getImages`
    """

    data: list[SandboxImageInfo] = []



from pydantic import BaseModel, Field


class SandboxClusterStatusResponse(BaseModel):
    """
    Cluster information
    """

    class Node(BaseModel):
        class Condition(BaseModel):
            type: str
            message: str

        class Info(BaseModel):
            """
            System Information: Component versions
            """

            kubernetes: str
            """
            Kubernetes version
            """

            container_runtime: str = Field(alias="containerRuntime")
            """
            Containerd version
            """

            os: str

            os_image: str = Field(alias="osImage")

            os_kernel: str = Field(alias="osKernel")

            xen_version: str = Field(alias="xenVersion")

            vm_cpu: int = Field(alias="vmCpu")
            """
            The number of cores allocated for the BA
            """

            vm_ram: int = Field(alias="vmRam")
            """
            The amount of RAM allocated for the BA in Bytes
            """

        class IPs(BaseModel):
            class IP(BaseModel):
                type: str
                ip: str

            internal: IP
            cluster: IP

        name: str

        hostname: str = Field(alias="hostName")

        ready: bool
        """
        Node status
        """

        cpu: int
        """
        Number of cores
        """

        ram: int
        """
        The amount of RAM in Bytes
        """

        vm_capacity: int = Field(alias="vmCapacity")
        """
        The number of traps per node. Total capacity.
        """

        vm_allocatable: int = Field(alias="vmAllocatable")
        """
        The number of traps per node. Currently in use.
        """

        conditions: list[Condition]
        """
        Problematic conditions on the node Problematic conditions on the node will be shown here.

        Examples of conditions:

        ```
        [
            { "message": "Calico is not running on this node", "type": "NetworkUnavailable" },
            { "message": "kubelet has unsufficient memory available", "type": "MemoryPressure" },
            { "message": "kubelet has disk pressure", "type": "DiskPressure" },
            { "message": "Kubelet stopped posting node status.", "type": "Ready" }
        ]
        ```
        """

        total_pods: int = Field(alias="totalPods")
        """
        The total number of pods per node
        """

        ready_pods: int = Field(alias="readyPods")

        running_pods: int | None = Field(default=None, alias="runningPods")
        """
        The number of working pods on the node
        """

        roles: list[str]
        """
        Node labels
        """

        info: Info

        ips: IPs

    high_availability: bool = Field(alias="highAvailability")
    """
    High availability mode
    """

    cluster_ip: str = Field(alias="clusterIp")

    sb_nodes: int = Field(alias="sbNodes")
    """
    Number of nodes
    """

    sb_nodes_available: int = Field(alias="sbNodesAvailable")
    """
    Number of available BA nodes
    """

    vms_count: int = Field(alias="vmsCount")
    """
    Number of VMs
    """

    vms_count_available: int = Field(alias="vmsCountAvailable")
    """
    Number of available VMs
    """

    nodes: list[Node] = []
    """
    List of nodes
    """



from pydantic import BaseModel, Field

from ptsandbox.models.core.enum import BaqueueState, EntryPointType


class SandboxBaqueueTasksResponse(BaseModel):
    """
    Listing of issues in the BA queue
    """

    class Task(BaseModel):
        task_id: str = Field(alias="taskId")

        correlation_id: str = Field(alias="correlationId")

        result_task_id: str | None = Field(default=None, alias="resultTaskId")
        """
        ID of the task from which the result was taken
        """

        order_number: int | None = Field(default=None, alias="orderNumber")
        """
        Sequence number in the queue (for unfinished tasks)
        """

        state: BaqueueState
        """
        Task status
        """

        priority: int
        """
        Task priority
        """

        priority_name: str = Field(alias="priorityName")
        """
        String representation of task priority
        """

        entry_point_id: str = Field(alias="entryPointId")
        """
        Source ID
        """

        entry_point_type: EntryPointType = Field(alias="entryPointType")
        """
        Source Type
        """

        object_mime_type: str = Field(alias="objectMimeType")
        """
        The object's mime type
        """

        task_object_name: str = Field(alias="taskObjectName")
        """
        Task Name
        """

        task_object_type: str = Field(alias="taskObjectType")
        """
        The type of the task object (FILE, EMAIL, ...)
        """

        object_properties: list[str] = Field(alias="objectProperties")
        """
        List of file labels
        """

        image_id: str | None = Field(default=None, alias="imageId")
        """
        ID of the BA image
        """

        planned_duration: int = Field(alias="plannedDuration")
        """
        Planned analysis time (seconds)
        """

        bootkitmon_enable: bool = Field(alias="bootkitmonEnable")
        """
        Is bootkitmon enabled or not
        """

        planned_bootkitmon_duration: int | None = Field(default=None, alias="plannedBootkitmonDuration")
        """
        Planned time of the second stage analysis (seconds)
        """

        planned_total_duration: float | None = Field(default=None, alias="plannedTotalDuration")
        """
        Scheduled total time (main + bootkit + costs) - seconds
        """

        real_duration: float | None = Field(default=None, alias="realDuration")
        """
        Actual duration (seconds)
        """

        object_name: str = Field(alias="objectName")
        """
        File Name
        """

        object_sha256: str = Field(alias="objectSha256")
        """
        sha256 file
        """

        object_sandbox_type: str = Field(alias="objectSandboxType")
        """
        Sandbox file type
        """

        save_video: bool = Field(alias="saveVideo")
        """
        Is video saving enabled or not
        """

        convert_video: bool = Field(alias="convertVideo")
        """
        Is video conversion enabled or not
        """

        procdump_enable: bool = Field(alias="procdumpEnable")
        """
        Is procdump enabled or not
        """

        custom_command: str = Field(alias="customCommand")
        """
        A custom command to run
        """

        ts_created: float = Field(alias="tsCreated")
        """
        Task creation timestamp
        """

        ts_starting: float | None = Field(default=None, alias="tsStarting")
        """
        Timestamp recruitment attempts
        """

        ts_started: float | None = Field(default=None, alias="tsStarted")
        """
        Timestamp of the start of work
        """

        ts_ready: float | None = Field(default=None, alias="tsReady")
        """
        Readiness timestamp
        """

        ts_finished: float | None = Field(default=None, alias="tsFinished")
        """
        Completion timestamp
        """

        estimated_ts_finished: float | None = Field(default=None, alias="estimatedTsFinished")
        """
        Predicted completion time
        """

        error_text: str | None = Field(default=None, alias="errorText")
        """
        The error text. In addition to the constants defined below, there may be any error text.

        TIMEOUT_START_EXCEEDED, TIMEOUT_FINISH_EXCEEDED, TIMEOUT_DOOMED_TO_START
        """

    total: int
    """
    The total number of tasks that meet the search conditions
    """

    tasks: list[Task]
    """
    Task list
    """



from ptsandbox.models.core.base import (
    BaseRequest,
    BaseResponse,
    SandboxException,
    SandboxFileNotFoundException,
    SandboxTooManyErrorsException,
    SandboxUploadException,
    SandboxWaitTimeoutException,
)
from ptsandbox.models.core.common import (
    Artifact,
    Detection,
    Log,
    NetworkObject,
    SandboxImageInfo,
    SandboxResult,
    SuspiciousBehaviors,
)
from ptsandbox.models.core.enum import (
    Action,
    ArtifactType,
    BaqueueState,
    BlacklistStatus,
    BootkitmonStage,
    ComponentStatus,
    ComponentType,
    ContextType,
    DeliveryStatus,
    DPIState,
    EmailType,
    EngineName,
    EngineSubsystem,
    EntryPointAction,
    EntryPointStatus,
    EntryPointType,
    EOSStatus,
    ErrorType,
    FileInfoProperties,
    FileInfoTypes,
    HashType,
    HTTPDirection,
    LicenseAvEngine,
    LicenseEntryPoint,
    LicensePerformanceType,
    LicenseStatus,
    LicenseUpdateError,
    LogType,
    NetworkObjectType,
    QuarantineEventType,
    QuarantineState,
    SandboxImageType,
    ScanArtifactType,
    ScanState,
    SystemCode,
    SystemGroup,
    ThreatClassification,
    ThreatPlatform,
    TokenPermissions,
    TreeEngineName,
    TreeNodeType,
    UploadStrategy,
    Verdict,
    VNCMode,
)

__all__ = [
    "Action",
    "Artifact",
    "ArtifactType",
    "BaqueueState",
    "BaseRequest",
    "BaseResponse",
    "BlacklistStatus",
    "BootkitmonStage",
    "ComponentStatus",
    "ComponentType",
    "ContextType",
    "DPIState",
    "DeliveryStatus",
    "Detection",
    "EOSStatus",
    "EmailType",
    "EngineName",
    "EngineSubsystem",
    "EntryPointAction",
    "EntryPointStatus",
    "EntryPointType",
    "ErrorType",
    "FileInfoProperties",
    "FileInfoTypes",
    "HTTPDirection",
    "HashType",
    "LicenseAvEngine",
    "LicenseEntryPoint",
    "LicensePerformanceType",
    "LicenseStatus",
    "LicenseUpdateError",
    "Log",
    "LogType",
    "NetworkObject",
    "NetworkObjectType",
    "QuarantineEventType",
    "QuarantineState",
    "SandboxException",
    "SandboxFileNotFoundException",
    "SandboxImageInfo",
    "SandboxImageType",
    "SandboxResult",
    "SandboxTooManyErrorsException",
    "SandboxUploadException",
    "SandboxWaitTimeoutException",
    "ScanArtifactType",
    "ScanState",
    "SuspiciousBehaviors",
    "SystemCode",
    "SystemGroup",
    "ThreatClassification",
    "ThreatPlatform",
    "TokenPermissions",
    "TreeEngineName",
    "TreeNodeType",
    "UploadStrategy",
    "VNCMode",
    "Verdict",
]



from dataclasses import dataclass
from enum import Enum
from typing import Any

import orjson
from loguru import logger


class DetectionType(str, Enum):
    SILENT = "silent"
    SUSPICIOUS = "suspicious"
    MALWARE = "malware"


@dataclass(frozen=True)
class Detect:
    name: str
    weight: int | None

    def __key(self) -> tuple[str]:
        return (self.name,)

    def __hash__(self) -> int:
        return hash(self.__key())

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Detect):
            return self.__key() == other.__key()  # noqa: W0212

        raise NotImplementedError()


class Detections:
    detections: dict[DetectionType, set[Detect]]

    _real_name: str

    def __init__(self, trace: bytes, ctx: str = "") -> None:
        self.detections = {
            DetectionType.SILENT: set(),
            DetectionType.SUSPICIOUS: set(),
            DetectionType.MALWARE: set(),
        }

        for line in trace.splitlines(keepends=False):
            try:
                event: dict[str, Any] = orjson.loads(line)
                if event.get("auxiliary.type", None) == "init":
                    self._real_name = event.get("object.name", "")

                detect_type: str = event.get("detect.type", "").upper()
                if detect_type in DetectionType._member_names_:
                    self.detections[DetectionType[detect_type]].add(
                        Detect(
                            name=event.get("detect.name", ""),
                            weight=event.get("weight", None),
                        )
                    )
            except Exception as ex:
                logger.error(f"Parsing trace exception: {ex!r} ({ctx = })")

    def __repr__(self) -> str:
        return repr(self.detections)

    @property
    def silent(self) -> set[Detect]:
        """Только silent-детекты"""
        return self.detections[DetectionType.SILENT]

    @property
    def suspicious(self) -> set[Detect]:
        """Только suspicious-детекты"""
        return self.detections[DetectionType.SUSPICIOUS]

    @property
    def malware(self) -> set[Detect]:
        """Только malware-детекты"""
        return self.detections[DetectionType.MALWARE]



from datetime import datetime
from typing import Literal

from pydantic import BaseModel, Field

from ptsandbox.models.core import (
    Action,
    DeliveryStatus,
    DPIState,
    EmailType,
    EntryPointAction,
    EntryPointStatus,
    EntryPointType,
    QuarantineEventType,
    ScanState,
    ThreatClassification,
    ThreatPlatform,
    TreeEngineName,
    Verdict,
)
from ptsandbox.models.core.enum import ErrorType, QuarantineState, TokenPermissions


class CorrelationInfo(BaseModel):
    """
    Information about correlation
    """

    state: ScanState
    """
    The state of correlation
    """

    threat_classification: ThreatClassification = Field(alias="threatClassification")
    """
    Object classification (VIRUS, SPAM, WORM, etc.)
    """

    threat_level: Verdict = Field(alias="threatLevel")
    """
    Threat level
    """

    threat_platform: ThreatPlatform = Field(alias="threatPlatform")
    """
    Artifact Platform
    """

    verdict_priority: int | None = Field(default=None, alias="verdictPriority")
    """
    Threat priority
    """


class MailResult(BaseModel):
    recipient: str
    action: Action
    email_type: EmailType = Field(alias="emailType")
    delivery_status: DeliveryStatus = Field(alias="deliveryStatus")
    server_address: str = Field(alias="serverAddress")
    server_port: int = Field(alias="serverPort")


class HTTPDescription(BaseModel):
    referer: str
    """
    The value of the HTTP 'Referer' header, from which page the request was sent
    """

    user_agent: str = Field(alias="userAgent")
    """
    The value of the HTTP header 'User-Agent'
    """

    host: str
    """
    The value of the HTTP header 'Host'
    """

    uri: str
    """
    Full request URL
    """


class EntryPoint(BaseModel):
    """
    Where did the task come from
    """

    class Quarantine(BaseModel):
        class QuarantineEvent(BaseModel):
            type: QuarantineEventType

            time: int
            """
            Event creation time (UNIX timestamp)
            """

            user_id: str | None = Field(default=None, alias="userId")
            """
            User ID (for SEND only)
            """

            smtp_host: str | None = Field(default=None, alias="smtpHost")
            """
            SMTP Host (for SEND only)
            """

            smtp_port: int | None = Field(default=None, alias="smtpPort")
            """
            SMTP Port (for SEND only)
            """

            recipients: list[str] | None = None
            """
            List of recipients (for SEND only)
            """

        state: QuarantineState
        """
        Quarantine state
        """

        events: list[QuarantineEvent] = []
        """
        List of quarantine events. Filled in only in API /summary, there is no such field in the listing.
        """

    class CheckMe(BaseModel):
        from_address: str = Field(alias="fromAddress")
        """
        The sender received from the SMTP session (the 'MAIL FROM' command)
        """

        recipients: list[str]
        """
        The list of recipients received from the SMTP session (the 'RCPT TO' command)
        """

    class ICAP(BaseModel):
        method: str
        """
        ICAP method (RESPMOD, REQMOD)
        """

        url: str
        """
        ICAP address of the service
        """

        version: str
        """
        ICAP version
        """

        client_ip: str = Field(alias="clientIp")
        """
        ICAP header value: 'X-Client-IP'
        """

        client_username: str = Field(alias="clientUsername")
        """
        ICAP header value: 'X-Client-Username'
        """

    class DPI(BaseModel):
        class SMTP(BaseModel):
            message_id: str = Field(..., alias="messageId")
            """
            The EML value of the 'Message-Id' header
            """

            sender: str
            """
            Sender of the received email header 'From'
            """

        src_ip: str = Field(alias="srcIp")
        """
        The IP address where the object was sent from
        """

        src_port: int = Field(alias="srcPort")
        """
        PORT where the object was sent from
        """

        dst_ip: str = Field(alias="dstIp")
        """
        The IP address where the object was sent to
        """

        dst_port: int = Field(..., alias="dstPort")
        """
        PORT where the object was sent to
        """

        proto: str
        """
        Protocol. For HTTP or SMTP values, the corresponding keys are added.
        """

        state: DPIState

        http: HTTPDescription | None = None

        smtp: SMTP | None = None

    class MailAgent(BaseModel):
        from_address: str = Field(alias="fromAddress")
        """
        The sender received from the SMTP session (the 'MAIL FROM' command)
        """

        recipients: list[str]
        """
        The list of recipients received from the SMTP session (the 'RCPT TO' command)
        """

        mail_results: list[MailResult] | None = Field(default=None, alias="mailResults")
        """
        The results are sent by mail. Filled in only in API /summary, there is no such field in the listing.
        """

    class MailBcc(BaseModel):
        from_address: str = Field(alias="fromAddress")
        """
        The sender received from the SMTP session (the 'MAIL FROM' command)
        """

        recipients: list[str]
        """
        The list of recipients received from the SMTP session (the 'RCPT TO' command)
        """

    class FileInbox(BaseModel):
        src_path: str = Field(alias="srcPath")
        """
        The original path to the file
        """

        dst_path: str = Field(alias="dstPath")
        """
        The path where the file was moved
        """

    class FileMonitor(BaseModel):
        src_path: str = Field(alias="srcPath")
        """
        The original path to the file
        """

    class MailGateway(BaseModel):
        from_address: str = Field(alias="fromAddress")
        """
        The sender received from the SMTP session (the 'MAIL FROM' command)
        """

        recipients: list[str]
        """
        The list of recipients received from the SMTP session (the 'RCPT TO' command)
        """

        mail_results: list[MailResult] | None = Field(default=None, alias="mailResults")
        """
        The results are sent by mail. Filled in only in API /summary, there is no such field in the listing.
        """

    class PTNAD(BaseModel):
        src_ip: str = Field(alias="srcIp")
        """
        The IP address where the object was sent from
        """

        src_port: int = Field(alias="srcPort")
        """
        PORT where the object was sent from
        """

        dst_ip: str = Field(..., alias="dstIp")
        """
        The IP address where the object was sent to
        """

        dst_port: int = Field(..., alias="dstPort")
        """
        PORT where the object was sent to
        """

        ref: str
        """
        Link to the PTNAD session
        """

        proto: str
        """
        Protocol
        """

        http: HTTPDescription | None = None

    class ClientWebInfo(BaseModel):
        user_agent: str = Field(alias="userAgent")
        """
        The value of the HTTP header 'User-Agent'
        """

        x_forwarded_for: str = Field(alias="xForwardedFor")
        """
        The value of the HTTP header 'X-Forwarded-For' is used to determine the IP of the HTTP client
        """

        referer: str
        """
        The value of the HTTP 'Referer' header, from which page the request was sent
        """

    class ClientFullWebInfo(ClientWebInfo):
        user_id: str = Field(alias="userId")

        user_login: str = Field(alias="userLogin")

        user_name: str = Field(alias="userName")

        user_is_anonymous: bool = Field(alias="userIsAnonymous")

    id: str
    """
    Source ID
    """

    type: EntryPointType
    """
    Source Type
    """

    status: EntryPointStatus
    """
    Completion status
    """

    action: EntryPointAction
    """
    Type of action
    """

    quarantine: Quarantine
    """
    Quarantine status
    """

    client_ip: str = Field(alias="clientIp")
    """
    The client's IP address
    """

    # TODO: check this field
    # metadata: dict[Any, Any]

    check_me: CheckMe | None = None
    """
    Information about the sender and recipients
    """

    icap: ICAP | None = None

    dpi: DPI | None = None

    mail_agent: MailAgent | None = Field(default=None, alias="mailAgent")

    mail_bcc: MailBcc | None = Field(default=None, alias="mailBcc")

    file_inbox: FileInbox | None = Field(default=None, alias="fileInbox")

    file_monitor: FileMonitor | None = Field(default=None, alias="fileMonitor")

    mail_gateway: MailGateway | None = Field(default=None, alias="mailGateway")

    mail_gateway_mta: MailGateway | None = Field(default=None, alias="mailGatewayMta")

    ptnad: PTNAD | None = None

    public_api: ClientWebInfo | None = Field(default=None, alias="publicApi")

    scan_api: ClientWebInfo | None = Field(default=None, alias="scanApi")

    pt_edr: ClientWebInfo | None = Field(default=None, alias="ptEdr")

    pt_cs: ClientWebInfo | None = Field(default=None, alias="ptCs")

    web: ClientFullWebInfo | None = None

    interactive_analysis: ClientFullWebInfo | None = Field(default=None, alias="interactiveAnalysis")


class DetectionUI(BaseModel):
    name: str
    threat_classification: ThreatClassification = Field(alias="threatClassification")
    id: str | None = None
    description: str | None = None


class Error(BaseModel):
    type: ErrorType

    duration: int | None = None
    """
    Waiting time
    """


class Scan(BaseModel):
    class Engine(BaseModel):
        name: TreeEngineName

        database_time: datetime | None = Field(default=None, alias="databaseTime")

        version: str

        detections: list[DetectionUI] = []

        errors: list[Error] = []

    engine: Engine

    result: CorrelationInfo | None = None


class FilterValues(BaseModel):
    class EntryPoint(BaseModel):
        name: str
        id: str
        type: EntryPointType

    entry_points: list[EntryPoint] = Field(alias="entryPoints")
    """
    Possible values for filters by source
    """

    threat_classifications: list[str] = Field(alias="threatClassifications")
    """
    Possible values for filters based on the analysis result
    """

    properties: list[str]
    """
    Возможные значения для фильтров по свойствам файла
    """

    categories: list[str]
    """
    Possible values for filters by link category
    """


class Token(BaseModel):
    class EntryPoint(BaseModel):
        """
        The entrypoint to which the api token is linked.

        It is used for listing and for obtaining a specific token.
        """

        id: str
        """
        Entrypoint ID
        """

        name: str
        """
        Entrypoint name
        """

    name: str
    """
    Name of the PublicAPI token

    pattern: ^[a-zA-Z][a-zA-Z0-9-]{3,28}[a-zA-Z]$
    """

    comment: str | None = None
    """
    Comment on the token
    """

    permissions: list[TokenPermissions]
    """
    Token Access Rights
    """

    id: int
    """
    ID of the record in the database
    """

    creator_login: str = Field(alias="creatorLogin")
    """
    Login of the user who created the token
    """

    created: float
    """
    UNIX time the time of token creation
    """

    modified: float | None = None
    """
    UNIX time the time when the token comment was changed
    """

    deleted: float | None = None
    """
    UNIX time the time of token deletion
    """

    entry_point: EntryPoint | None = Field(default=None, alias="entryPoint")


class SMTPDefaultRecord(BaseModel):
    """
    SMTP connection settings
    """

    class Config(BaseModel):
        class Credential(BaseModel):
            login: str
            password: str
            id: str

        class Timeout(BaseModel):
            connect: int
            execute: int

        host: str
        port: int
        use_ssl: bool = Field(alias="useSsl")
        auth_type: Literal["any", "plain", "encrypt", "ntlm", "none"] = Field(alias="authType")
        timeout: Timeout | None = None
        credential: Credential | None = None

    config: Config

    priority: int
    """
    Priority for recording
    """



from typing import NotRequired
from uuid import UUID

from pydantic import BaseModel, Field
from typing_extensions import TypedDict

from ptsandbox.models.core import (
    Artifact,
    BaseRequest,
    BaseResponse,
    SandboxException,
    SandboxResult,
    VNCMode,
)
from ptsandbox.models.core.enum import FileInfoProperties


class DebugOptions(TypedDict):
    """
    Description of all available debugging options for very detailed scan configuration
    """

    keep_sandbox: NotRequired[bool]
    """
    Don't destroy the sandbox after scanning
    """

    skip_work: NotRequired[bool]
    """
    Perform a scan, skipping the data collection stage for analysis
    """

    extract_crashdumps: NotRequired[bool]
    """
    Extract crashdumps from the sandbox
    """

    save_debug_files: NotRequired[bool]
    """
    Save files necessary for debugging (error logs, tcpdump logs, etc)
    """

    rules_url: NotRequired[str]
    """
    Use the specified normalization and correlation rules
    The rules are specified as a link to the archive containing the compiled rules
    """

    sleep_work: NotRequired[bool]
    """
    Perform a scan, replacing the data collection stage for analysis with an equivalent waiting time
    """

    disable_syscall_hooks: NotRequired[bool]
    """
    Disable syscall hooks functionality

    Read more about these hooks in documentation
    """

    disable_dll_hooks: NotRequired[bool]
    """
    Disable dll hooks functionality

    Read more about these hooks in documentation
    """

    custom_syscall_hooks: NotRequired[str]
    """
    Use the specified list of system calls to intercept

    The list is transmitted as an http link to a file with the names of system calls

    Read more about this file in [documentation](https://security-experts-community.github.io/py-ptsandbox/usage/public-api/scanning/debug-options/#custom_syscall_hooks)
    """

    custom_dll_hooks: NotRequired[str]
    """
    Use the specified list of system calls to intercept

    The list is transmitted as an http link to a file with the names of dll hooks for apimon plugin

    Read more about this file in [documentation](https://security-experts-community.github.io/py-ptsandbox/usage/public-api/scanning/debug-options/#custom_dll_hooks)
    """

    disable_retries: NotRequired[bool]
    """
    Disable task re-execution in case of a scan error
    """

    enable_sanitizers: NotRequired[bool]
    """
    Enable the debugging mechanisms of the sanitizers group
    """

    allowed_outbound_connections: NotRequired[list[str]]
    """
    Whitelist of IP addresses to which connections from a VM are allowed (backconnect)
    """

    payload_completion_event: NotRequired[str]
    """
    A regular expression for the raw DRAKVUF event, signaling the end of the useful work of the sample.

    If this option is specified, sandbox-worker will calculate and log the PAYLOAD_SCAN_TIME metric.
    """

    disable_procdump_on_finish: NotRequired[bool]
    """
    Disable the functionality of removing the memory dump from the sample at the end of the observation
    """

    skip_update_time: NotRequired[bool]
    """
    Do not synchronize the time in the VM with the host
    """

    disable_manual_scan_events: NotRequired[bool]
    """
    Do not send lifecycle notifications for manual behavioral analysis (console is ready, console is closed, etc.)
    """

    bootkitmon_boot_timeout: NotRequired[int]
    """
    The maximum waiting time for VM loading in seconds (90 seconds by default)
    """

    custom_procdump_exclude: NotRequired[str]
    """
    A file with a list of processes for which memory dumps should not be removed.

    Each line in the file is a regular expression of the path to the process file on disk.

    Read more about this file in [documentation](https://security-experts-community.github.io/py-ptsandbox/usage/public-api/scanning/debug-options/#custom_procdump_exclude)
    """

    custom_fileextractor_exclude: NotRequired[str]
    """
    A file with a list of files that should not be extracted

    Each line in the file is a regular expression of the path to the file on disk.

    Read more about this file in [documentation](https://security-experts-community.github.io/py-ptsandbox/usage/public-api/scanning/debug-options/#custom_fileextractor_exclude)
    """

    validate_plugins: NotRequired[bool]
    """
    Check plugins for at least one event during the entire behavioral analysis
    """

    extra_vm_init_url: NotRequired[str]
    """
    Run this script in the VM immediately before launching the behavioral analysis.

    It is useful, for example, to check the network during analysis.
    """


class SandboxOptions(BaseRequest):
    """
    Parameters of behavioral analysis.

    In the absence, the source parameters are used for analysis, which are set in the system by default.
    """

    class FilterProperties(BaseModel):
        """
        Filtering a group of files by properties to send to the sandbox for analysis
        """

        pdf: list[FileInfoProperties] = []

        office: list[FileInfoProperties] = []

    enabled: bool = True
    """
    Perform a behavioral analysis
    """

    image_id: str = "win7-sp1-x64"
    """
    ID of the VM image.

    You can view it in the sandbox interface.
    """

    custom_command: str | None = None
    """
    The command to run the file.

    The `{file}` marker in the string is replaced with the path to the file.

    For example: `rundll32.exe {file},#1`
    """

    procdump_new_processes_on_finish: bool = False
    """
    Take dumps for all spawned and non-dead processes
    """

    analysis_duration: int = Field(default=120, ge=10)
    """
    The duration of analysis the file in seconds. minimum: 10
    """

    bootkitmon: bool = False
    """
    Perform bootkitmon analysis
    """

    analysis_duration_bootkitmon: int = Field(default=60, ge=10)
    """
    The duration of analysis at the bootkitmon stage in seconds. minimum: 10
    """

    save_video: bool = True
    """
    Save video capture of the screen
    """

    mitm_enabled: bool = True
    """
    Enable certificates injection with PT Sandbox certificates when decrypting and analyzing secure traffic
    """

    file_types: list[str] | None = None
    """
    A list of the final file types or groups of files that will be sent for behavioral analysis

    For example:
    ["adobe-acrobat/", "databases/", "executable-files/", "presentations/", "spreadsheets/", "word-processor/"]
    """

    filter_by_properties: FilterProperties | None = None
    """
    Filtering a group of files by properties to send to the sandbox for analysis
    """

    debug_options: DebugOptions = {"save_debug_files": False}
    """
    Fine-tuning
    """


class SandboxOptionsAdvanced(BaseRequest):
    """
    Run an advanced analysis of the uploaded file in the VM without unpacking.

    Provides an opportunity to fine-tuning.

    **The options are in beta, so they may change in the future.**
    """

    class ExtraFile(BaseModel):
        """
        An additional file to be placed next to the sample
        """

        uri: str
        """
        Link to the uploaded object
        """

        name: str
        """
        Name in the VM
        """

    image_id: str = "win7-sp1-x64"
    """
    ID of the VM image.

    You can view it in the sandbox interface.
    """

    custom_command: str | None = None
    """
    The command to run the file.

    The `{file}` marker in the string is replaced with the path to the file.

    For example: `rundll32.exe {file},#1`
    """

    procdump_new_processes_on_finish: bool = False
    """
    Take dumps for all spawned and non-dead processes
    """

    analysis_duration: int = Field(default=120, ge=10)
    """
    The duration of analysis the file in seconds. minimum: 10
    """

    bootkitmon: bool = False
    """
    Perform bootkitmon analysis
    """

    analysis_duration_bootkitmon: int = Field(default=60, ge=10)
    """
    The duration of analysis at the bootkitmon stage in seconds. minimum: 10
    """

    save_video: bool = True
    """
    Save video capture of the screen
    """

    mitm_enabled: bool = True
    """
    Enable certificates injection with PT Sandbox certificates when decrypting and analyzing secure traffic
    """

    disable_clicker: bool = False
    """
    Disable auto-clicker startup

    Useful when enabling manual analysis.
    """

    skip_sample_run: bool = False
    """
    Disable sample launch
    """

    vnc_mode: VNCMode = VNCMode.DISABLED
    """
    Manual analysis mode
    """

    extra_files: list[ExtraFile] = []
    """
    A list of additional files that are placed in the VM
    """

    debug_options: DebugOptions = {"save_debug_files": False}
    """
    Fine-tuning
    """


class SandboxBaseScanTaskRequest(BaseRequest):
    """
    Base class for all scan requests
    """

    class Options(BaseModel):
        class SuspiciousFilesOptions(BaseModel):
            """
            Settings for marking files as suspicious
            """

            encrypted_not_unpacked: bool = True
            """
            Encrypted and not unpacked file
            """

            max_depth_exceeded: bool = True
            """
            Unpacking depth exceeded
            """

            office_encrypted: bool = True
            """
            Encrypted office file
            """

            office_has_macros: bool = True
            """
            Office file with macros
            """

            office_has_embedded: bool = True
            """
            Office file with embedded objects
            """

            office_has_active_x: bool = True
            """
            Office file with ActiveX controls
            """

            office_has_dde: bool = True
            """
            Office file with dynamic data exchange
            """

            office_has_remote_data: bool = True
            """
            Office file with remote data
            """

            office_has_remote_template: bool = True
            """
            Office file with remote templates
            """

            office_has_action: bool = True
            """
            Office file with Action
            """

            pdf_encrypted: bool = True
            """
            Encrypted PDF file
            """

            pdf_has_embedded: bool = True
            """
            PDF file with embedded objects
            """

            pdf_has_open_action: bool = True
            """
            PDF file with Open Action
            """

            pdf_has_action: bool = True
            """
            PDF file with Action
            """

            pdf_has_javascript: bool = True
            """
            PDF file with Javascript
            """

            pdf_protected: bool = True
            """
            Protected PDF file
            """

        analysis_depth: int = 0
        """
        The depth of the check.

        The maximum level of decomposition of objects with a hierarchical structure (archives, emails, links, etc.)
        or the decompression level of compressed files.

        If the value is 0, the check is performed without decomposition and decompression.

        The higher the number, the longer the check can take.
        """

        scan_timeout: int = Field(default=1200, ge=10, le=3600)
        """
        Maximum response time
        """

        max_execution_time: int = Field(default=3600, ge=300, le=10800)
        """
        Maximum waiting time for analysis
        """

        passwords_for_unpack: list[str] = []
        """
        List of passwords for unpacking encrypted archives
        """

        cache_enabled: bool = False
        """
        If the file has already been analyzed before, it will be taken from the cache, and not analyzed again.
        """

        url_extract_enabled: bool = True
        """
        Extract links from objects
        """

        enable_experimental_yara_rules: bool = False
        """
        Enable object analysis using yara test rules
        """

        mark_suspicious_files_options: SuspiciousFilesOptions | None = None
        """
        Settings for marking files as suspicious. By default, we do not send, but take the settings from the sandbox.

        You can configure it by passing an object with the necessary options.
        """

        sandbox: SandboxOptions = SandboxOptions()
        """
        Behavioral Analysis Parameters
        """

    file_name: str | None = None
    """
    The name of the file to be checked, which will be displayed in the sandbox web interface.

    If not specified, the hash value of the file is calculated using the SHA—256 algorithm.
    """

    short_result: bool = False
    """
    Return only the overall result of the check.

    The parameter value is ignored (true is used) if the value of the `async_result` parameter is also `true`.
    """

    async_result: bool = True
    """
    Return only the scan_id.

    Enabling this option may be usefull to send async requests for file checking.

    You can receive full report in a separate request.
    """

    priority: int = Field(default=3, ge=1, le=4)
    """
    The priority of the task. The higher it is, the faster it will get to work
    """


class SandboxScanTaskRequest(SandboxBaseScanTaskRequest):
    """
    Parameters of an API request to start analyzing a file previously uploaded to the product.

    `<URL>/analysis/createScanTask`
    """

    file_uri: str
    """
    The file URI received when uploading the file
    """

    options: SandboxBaseScanTaskRequest.Options = SandboxBaseScanTaskRequest.Options()
    """
    Additional scanning options
    """


class SandboxAdvancedScanTaskRequest(SandboxBaseScanTaskRequest):
    """
    Parameters of an API request to start analyzing a file previously uploaded to the product.

    `<URL>/analysis/createBAScanTask`
    """

    file_uri: str
    """
    The file URI received when uploading the file
    """

    sandbox: SandboxOptionsAdvanced = SandboxOptionsAdvanced()
    """
    Additional advanced scanning options
    """


class SandboxRescanTaskRequest(SandboxBaseScanTaskRequest):
    """
    API request parameters for launching retro analysis.

    Allows scanning with the new drakvuf-trace.log.zst and tcpdump.pcap rules

    `<URL>/analysis/createRetroTask`
    """

    file_uri: str
    """
    The file URI received when uploading the file
    """

    raw_events_uri: str | None = None
    """
    Temporary URI of the raw trace file
    """

    raw_network_uri: str | None = None
    """
    Temporary URI of the network file
    """

    options: SandboxBaseScanTaskRequest.Options = SandboxBaseScanTaskRequest.Options()
    """
    Additional scanning options
    """


class SandboxScanURLTaskRequest(SandboxBaseScanTaskRequest):
    """
    Parameters of the API request to start URL analysis.

    `<URL>/analysis/createScanURLTask`
    """

    url: str
    """
    URL address for analysis
    """

    options: SandboxBaseScanTaskRequest.Options = SandboxBaseScanTaskRequest.Options()
    """
    Additional scanning options
    """


class SandboxBaseTaskResponse(BaseResponse):
    """
    Base class for all scan responses
    """

    class ShortReport(BaseModel):
        scan_id: UUID
        """
        ID of the created task
        """

    class LongReport(ShortReport):
        result: SandboxResult
        """
        The overall result of the check.

        Missing from search responses:
            * `createScanTask` with the `async_result` parameter enabled;
            * `checkTask`, if the file analysis has not been completed yet
        """

        artifacts: list[Artifact]
        """
        A file, email, or other object that was checked during file analysis.

        Missing from search responses:
            * `createScanTask` with the `async_result` or `short_result` option enabled;
            * `checkTask`
        """

    data: LongReport | ShortReport = Field(union_mode="left_to_right")
    """
    Only the ShortReport is returned if async_result = True
    """

    def get_short_report(self) -> ShortReport:
        if self.errors:
            raise SandboxException(f"{self.errors}")

        return self.data

    def get_long_report(self) -> LongReport | None:
        if self.errors:
            raise SandboxException(f"{self.errors}")

        if isinstance(self.data, SandboxBaseTaskResponse.LongReport):
            return self.data

        return None


class SandboxCheckTaskRequest(BaseRequest):
    """
    Parameters of the API request for receiving file analysis results.

    The request can be used to get the results of the file analysis,
    which was started by an asynchronous request (`createScanTask` with the `async_result` parameter enabled).

    `<URL>/analysis/checkTask`
    """

    scan_id: UUID
    """
    ID of the task
    """

    allow_preflight: bool = True
    """
    If this flag is set, an intermediate result with the `is_preflight` attribute
    will be returned for scanning with multiple stages (for example, static + BA).
    """


class SandboxCheckTaskResponse(BaseResponse):
    class Data(BaseModel):
        scan_id: UUID
        """
        ID of the created task
        """

        result: SandboxResult | None = None
        """
        The overall result of the check.

        Missing from search responses:
            * `createScanTask` with the `async_result` parameter enabled;
            * `checkTask`, if the file analysis has not been completed yet
        """

        is_preflight: bool
        """
        Is the result preliminary, for example, only static has completed
        """

    data: Data



from typing import Any

import aiohttp
import aiohttp.client_exceptions
from loguru import logger
from pydantic import BaseModel, ConfigDict, ValidationError
from typing_extensions import Self


class BaseRequest(BaseModel):
    """
    The base class for all Request models related to the sandbox
    """

    model_config = ConfigDict(use_enum_values=True)

    def dict(self, exclude_none: bool = True, by_alias: bool = True, **kwargs: Any) -> dict[str, Any]:
        # The API does not like fields with None, so they must be excluded before exporting.
        return super().model_dump(exclude_none=exclude_none, by_alias=by_alias, **kwargs)

    def json(self, exclude_none: bool = True, by_alias: bool = True, **kwargs: Any) -> str:
        # The API does not like fields with None, so they must be excluded before exporting.
        return super().model_dump_json(exclude_none=exclude_none, by_alias=by_alias, **kwargs)


class BaseResponse(BaseModel):
    """
    The base class for all Response models related to the sandbox
    """

    class Error(BaseModel):
        message: str

        type: str

    data: Any
    errors: list[Error]

    @classmethod
    async def build(cls, response: aiohttp.ClientResponse) -> Self:
        try:
            return cls.model_validate(await response.json())
        except (ValidationError, aiohttp.client_exceptions.ContentTypeError) as err:
            logger.error(f"error: {err}")
            raise err


class SandboxException(Exception): ...


class SandboxUploadException(SandboxException): ...


class SandboxTooManyErrorsException(SandboxException): ...


class SandboxWaitTimeoutException(SandboxException): ...


class SandboxFileNotFoundException(SandboxException): ...
